"""Tests for the 'ISO3' categorization."""

import pytest

import climate_categories


def test_categories():
    assert climate_categories.ISO3["DEU"].title == "Germany"
    assert len(climate_categories.ISO3) > 150


def test_eu():
    assert len(climate_categories.ISO3["EU-12"].children[0]) == 12
    assert len(climate_categories.ISO3["EU-15"].children[0]) == 15
    assert len(climate_categories.ISO3["EU-25"].children[0]) == 25
    assert len(climate_categories.ISO3["EU-27_2007"].children[0]) == 27
    assert len(climate_categories.ISO3["EU-28"].children[0]) == 28
    assert len(climate_categories.ISO3["EU-27_2020"].children[0]) == 27
    with pytest.raises(KeyError):
        climate_categories.ISO3["EU27"]

    assert climate_categories.ISO3["EU"] == climate_categories.ISO3["EU_2020"]


def test_unfccc():
    assert len(climate_categories.ISO3["Annex-I"].children[0]) == 43
    assert len(climate_categories.ISO3["Non-Annex-I"].children[0]) == 155
    # leaf children excludes EU, because EU is not a leaf
    assert len(climate_categories.ISO3["UNFCCC"].leaf_children[0]) == 197
    assert len(climate_categories.ISO3["UNFCCC"].children[0]) == 198
    assert (
        climate_categories.ISO3["EU"] in climate_categories.ISO3["UNFCCC"].children[0]
    )
    assert climate_categories.ISO3["Annex-I"] in climate_categories.ISO3.descendants(
        "UNFCCC"
    )
    assert climate_categories.ISO3[
        "Non-Annex-I"
    ] in climate_categories.ISO3.descendants("UNFCCC")


def test_g7g20():
    # since the EU is a non-enumerated member, G7 has 8 members, G8 has 9.
    assert len(climate_categories.ISO3["G7"].children[0]) == 8
    assert len(climate_categories.ISO3["G8"].children[0]) == 9
    # in the G20, the EU is enumerated
    assert len(climate_categories.ISO3["G20"].children[0]) == 20


def test_oecd():
    assert len(climate_categories.ISO3["OECD"].leaf_children[0]) == 38


def test_aosis():
    assert len(climate_categories.ISO3["AOSIS"].leaf_children[0]) == 39


def test_gcam():
    assert (
        len(climate_categories.ISO3_GCAM["GCAM 7.0|Southeast Asia"].leaf_children[0])
        == 37
    )



import climate_categories


def test_number_of_categories():
    assert (
        len(climate_categories.IPCC2006_PRIMAP) == len(climate_categories.IPCC2006) + 20
    )


def test_levels():
    assert climate_categories.IPCC2006_PRIMAP.level("1") == 2
    assert climate_categories.IPCC2006_PRIMAP.level("1.A.1.bc") == 5


def test_consistent():
    for cat in climate_categories.IPCC2006_PRIMAP.values():
        if "corresponding_categories_IPCC1996" in cat.info:
            for ccat in cat.info["corresponding_categories_IPCC1996"]:
                assert ccat in climate_categories.IPCC1996



# Do not edit this file. It was auto-generated from the
# corresponding YAML file.
spec = {
    "name": "RCMIP",
    "title": "Emissions categories from the Reduced Complexity Model Intercomparison Project (RCMIP)",
    "comment": "AFOLU in the SSPDb is AFOLU minus any agriculture\nrelated fossil fuel based emissions hence is not the same as the\nWG3 AFOLU definition. Rather AFOLU in the SSPDb is AFOLU as expected by\nMAGICC (i.e. exluding agriculture related fossil fuel use), hence\nwe call it MAGICC AFOLU.",
    "references": "Nicholls, Z. R. J., Meinshausen, M., Lewis, J., Gieseke, R., Dommenget, D., Dorheim, K., Fan, C.-S., Fuglestvedt, J. S., Gasser, T., Golüke, U., Goodwin, P., Hartin, C., Hope, A. P., Kriegler, E., Leach, N. J., Marchegiani, D., McBride, L. A., Quilcaille, Y., Rogelj, J., Salawitch, R. J., Samset, B. H., Sandstad, M., Shiklomanov, A. N., Skeie, R. B., Smith, C. J., Smith, S., Tanaka, K., Tsutsui, J., and Xie, Z.: Reduced Complexity Model Intercomparison Project Phase 1: introduction and evaluation of global-mean temperature response, Geosci. Model Dev., 13, 5175-5190, https://doi.org/10.5194/gmd-13-5175-2020, 2020.",
    "institution": "RCMIP",
    "hierarchical": True,
    "last_update": "2020-09-21",
    "version": "v5.1.0",
    "total_sum": True,
    "canonical_top_level_category": "Emissions",
    "categories": {
        "Emissions": {
            "title": "RCMIP Emissions",
            "children": [
                [
                    "Emissions|BC",
                    "Emissions|CH4",
                    "Emissions|CO",
                    "Emissions|CO2",
                    "Emissions|F-Gases",
                    "Emissions|Montreal Gases",
                    "Emissions|N2O",
                    "Emissions|NH3",
                    "Emissions|NOx",
                    "Emissions|OC",
                    "Emissions|Sulfur",
                    "Emissions|VOC",
                ]
            ],
        },
        "Emissions|BC": {
            "title": "Black carbon emissions",
            "comment": "total black carbon emissions",
            "children": [
                [
                    "Emissions|BC|MAGICC AFOLU",
                    "Emissions|BC|MAGICC Fossil and Industrial",
                    "Emissions|BC|Other",
                ]
            ],
        },
        "Emissions|BC|MAGICC AFOLU": {
            "title": "Black carbon AFOLU emissions",
            "comment": "black carbon emissions from agriculture, forestry and other land use (IPCC category 3), excluding any fossil-fuel based emissions in the Agricultural sector (hence not identical to WG3 AFOLU)",
        },
        "Emissions|BC|MAGICC Fossil and Industrial": {
            "title": "Black carbon fossil and industrial emissions",
            "comment": "black carbon emissions from energy use on supply and demand side (IPCC category 1A, 1B), industrial processes (IPCC category 2), waste (IPCC category 4) and other (IPCC category 5)",
        },
        "Emissions|BC|Other": {
            "title": "Black carbon emissions from other sources",
            "comment": "black carbon emissions from other sources (please provide a definition)",
        },
        "Emissions|CH4": {
            "title": "Methane emissions",
            "comment": "total methane emissions",
            "children": [
                [
                    "Emissions|CH4|MAGICC AFOLU",
                    "Emissions|CH4|MAGICC Fossil and Industrial",
                    "Emissions|CH4|Other",
                ]
            ],
        },
        "Emissions|CH4|MAGICC AFOLU": {
            "title": "Methane AFOLU emissions",
            "comment": "methane emissions from agriculture, forestry and other land use (IPCC category 3), excluding any fossil-fuel based emissions in the Agricultural sector (hence not identical to WG3 AFOLU)",
        },
        "Emissions|CH4|MAGICC Fossil and Industrial": {
            "title": "Methane fossil and industrial emissions",
            "comment": "methane emissions from energy use on supply and demand side (IPCC category 1A, 1B), industrial processes (IPCC category 2), waste (IPCC category 4) and other (IPCC category 5)",
        },
        "Emissions|CH4|Other": {
            "title": "Methane emissions from other sources",
            "comment": "methane emissions from other sources (please provide a definition)",
        },
        "Emissions|CO": {
            "title": "Carbon monoxide emissions",
            "comment": "total carbon monoxide emissions",
            "children": [
                [
                    "Emissions|CO|MAGICC AFOLU",
                    "Emissions|CO|MAGICC Fossil and Industrial",
                    "Emissions|CO|Other",
                ]
            ],
        },
        "Emissions|CO|MAGICC AFOLU": {
            "title": "Carbon monoxide AFOLU emissions",
            "comment": "carbon monoxide emissions from agriculture, forestry and other land use (IPCC category 3), excluding any fossil-fuel based emissions in the Agricultural sector (hence not identical to WG3 AFOLU)",
        },
        "Emissions|CO|MAGICC Fossil and Industrial": {
            "title": "Carbon monoxide fossil and industrial emissions",
            "comment": "carbon monoxide emissions from energy use on supply and demand side (IPCC category 1A, 1B), industrial processes (IPCC category 2), waste (IPCC category 4) and other (IPCC category 5)",
        },
        "Emissions|CO|Other": {
            "title": "Carbon monoxide emissions from other sources",
            "comment": "carbon monoxide emissions from other sources (please provide a definition)",
        },
        "Emissions|CO2": {
            "title": "Carbon dioxide emissions",
            "comment": "total carbon dioxide emissions",
            "children": [
                [
                    "Emissions|CO2|MAGICC AFOLU",
                    "Emissions|CO2|MAGICC Fossil and Industrial",
                    "Emissions|CO2|Other",
                ]
            ],
        },
        "Emissions|CO2|MAGICC AFOLU": {
            "title": "Carbon dioxide AFOLU emissions",
            "comment": "carbon dioxide emissions from agriculture, forestry and other land use (IPCC category 3), excluding any fossil-fuel based emissions in the Agricultural sector (hence not identical to WG3 AFOLU)",
        },
        "Emissions|CO2|MAGICC Fossil and Industrial": {
            "title": "Carbon dioxide fossil and industrial emissions",
            "comment": "carbon dioxide emissions from energy use on supply and demand side (IPCC category 1A, 1B), industrial processes (IPCC category 2), waste (IPCC category 4) and other (IPCC category 5)",
        },
        "Emissions|CO2|Other": {
            "title": "Carbon dioxide emissions from other sources",
            "comment": "carbon dioxide emissions from other sources (please provide a definition)",
        },
        "Emissions|F-Gases": {
            "title": "F-gas emissions",
            "comment": "total F-gas emissions, including sulfur hexafluoride (SF6), nitrogen trifluoride (NF3), sulfuryl fluoride (SO2F2), hydrofluorocarbons (HFCs), perfluorocarbons (PFCs) ",
            "children": [
                [
                    "Emissions|F-Gases|HFC",
                    "Emissions|F-Gases|NF3",
                    "Emissions|F-Gases|PFC",
                    "Emissions|F-Gases|SF6",
                    "Emissions|F-Gases|SO2F2",
                ]
            ],
        },
        "Emissions|F-Gases|HFC": {
            "title": "Hydrofluorocarbons (HFCs and HCFCs) emissions",
            "comment": "equivalent species total emissions of hydrofluorocarbons (HFCs and HCFCs), provided as aggregate CO2-equivalents",
            "children": [
                [
                    "Emissions|F-Gases|HFC|HFC125",
                    "Emissions|F-Gases|HFC|HFC134a",
                    "Emissions|F-Gases|HFC|HFC143a",
                    "Emissions|F-Gases|HFC|HFC152a",
                    "Emissions|F-Gases|HFC|HFC227ea",
                    "Emissions|F-Gases|HFC|HFC23",
                    "Emissions|F-Gases|HFC|HFC236fa",
                    "Emissions|F-Gases|HFC|HFC245fa",
                    "Emissions|F-Gases|HFC|HFC32",
                    "Emissions|F-Gases|HFC|HFC365mfc",
                    "Emissions|F-Gases|HFC|HFC4310mee",
                ]
            ],
        },
        "Emissions|F-Gases|HFC|HFC125": {
            "title": "HFC125 emissions",
            "comment": "total emissions of HFC125",
        },
        "Emissions|F-Gases|HFC|HFC134a": {
            "title": "HFC134a emissions",
            "comment": "total emissions of HFC134a",
        },
        "Emissions|F-Gases|HFC|HFC143a": {
            "title": "HFC143a emissions",
            "comment": "total emissions of HFC143a",
        },
        "Emissions|F-Gases|HFC|HFC152a": {
            "title": "HFC152a emissions",
            "comment": "total emissions of HFC152a",
        },
        "Emissions|F-Gases|HFC|HFC227ea": {
            "title": "HFC227ea emissions",
            "comment": "total emissions of HFC227ea",
        },
        "Emissions|F-Gases|HFC|HFC23": {
            "title": "HFC23 emissions",
            "comment": "total emissions of HFC23",
        },
        "Emissions|F-Gases|HFC|HFC236fa": {
            "title": "HFC236fa emissions",
            "comment": "total emissions of HFC236fa",
        },
        "Emissions|F-Gases|HFC|HFC245fa": {
            "title": "HFC245fa emissions",
            "comment": "total emissions of HFC245fa",
        },
        "Emissions|F-Gases|HFC|HFC32": {
            "title": "HFC32 emissions",
            "comment": "total emissions of HFC32",
        },
        "Emissions|F-Gases|HFC|HFC365mfc": {
            "title": "HFC365mfc emissions",
            "comment": "total emissions of HFC365mfc",
        },
        "Emissions|F-Gases|HFC|HFC4310mee": {
            "title": "HFC43-10mee emissions",
            "comment": "total emissions of HFC43-10mee",
        },
        "Emissions|F-Gases|NF3": {
            "title": "Nitrogen trifluoride emissions",
            "comment": "total emissions of nitrogen trifluoride (NF3)",
        },
        "Emissions|F-Gases|PFC": {
            "title": "Perfluorocarbons  emissions",
            "comment": "equivalent species total emissions of perfluorocarbons (PFCs, as defined by Table 8.A.1 of AR5), provided as aggregate CF4-equivalents",
            "children": [
                [
                    "Emissions|F-Gases|PFC|C2F6",
                    "Emissions|F-Gases|PFC|C3F8",
                    "Emissions|F-Gases|PFC|C4F10",
                    "Emissions|F-Gases|PFC|C5F12",
                    "Emissions|F-Gases|PFC|C6F14",
                    "Emissions|F-Gases|PFC|C7F16",
                    "Emissions|F-Gases|PFC|C8F18",
                    "Emissions|F-Gases|PFC|CF4",
                    "Emissions|F-Gases|PFC|cC4F8",
                ]
            ],
        },
        "Emissions|F-Gases|PFC|C2F6": {
            "title": "C2F6 emissions",
            "comment": "total emissions of C2F6",
        },
        "Emissions|F-Gases|PFC|C3F8": {
            "title": "C3F8 emissions",
            "comment": "total emissions of C3F8",
        },
        "Emissions|F-Gases|PFC|C4F10": {
            "title": "C4F10 emissions",
            "comment": "total emissions of C4F10",
        },
        "Emissions|F-Gases|PFC|C5F12": {
            "title": "C5F12 emissions",
            "comment": "total emissions of C5F12",
        },
        "Emissions|F-Gases|PFC|C6F14": {
            "title": "C6F14 emissions",
            "comment": "total emissions of C6F14",
        },
        "Emissions|F-Gases|PFC|C7F16": {
            "title": "C7F16 emissions",
            "comment": "total emissions of C7F16",
        },
        "Emissions|F-Gases|PFC|C8F18": {
            "title": "C8F18 emissions",
            "comment": "total emissions of C8F18",
        },
        "Emissions|F-Gases|PFC|cC4F8": {
            "title": "c-C4F8 emissions",
            "comment": "total emissions of c-C4F8",
        },
        "Emissions|F-Gases|PFC|CF4": {
            "title": "CF4 emissions",
            "comment": "total emissions of CF4",
        },
        "Emissions|F-Gases|SF6": {
            "title": "Sulfur hexafluoride emissions",
            "comment": "total emissions of sulfur hexafluoride (SF6)",
        },
        "Emissions|F-Gases|SO2F2": {
            "title": "Sulfuryl fluoride emissions",
            "comment": "total emissions of sulfuryl fluoride (SO2F2)",
        },
        "Emissions|Montreal Gases": {
            "title": "Montreal gas emissions",
            "comment": "equivalent species total Montreal gas emissions, provided as CFC-11 equivalents",
            "children": [
                [
                    "Emissions|Montreal Gases|CCl4",
                    "Emissions|Montreal Gases|CFC",
                    "Emissions|Montreal Gases|CH2Cl2",
                    "Emissions|Montreal Gases|CH3Br",
                    "Emissions|Montreal Gases|CH3CCl3",
                    "Emissions|Montreal Gases|CH3Cl",
                    "Emissions|Montreal Gases|CHCl3",
                    "Emissions|Montreal Gases|HCFC141b",
                    "Emissions|Montreal Gases|HCFC142b",
                    "Emissions|Montreal Gases|HCFC22",
                    "Emissions|Montreal Gases|Halon1202",
                    "Emissions|Montreal Gases|Halon1211",
                    "Emissions|Montreal Gases|Halon1301",
                    "Emissions|Montreal Gases|Halon2402",
                ]
            ],
        },
        "Emissions|Montreal Gases|CCl4": {
            "title": "CCl4 emissions",
            "comment": "total emissions of CCl4",
        },
        "Emissions|Montreal Gases|CFC": {
            "title": "CFC emissions",
            "comment": "equivalent species total CFC emissions, provided as CFC-11 equivalents",
            "children": [
                [
                    "Emissions|Montreal Gases|CFC|CFC11",
                    "Emissions|Montreal Gases|CFC|CFC113",
                    "Emissions|Montreal Gases|CFC|CFC114",
                    "Emissions|Montreal Gases|CFC|CFC115",
                    "Emissions|Montreal Gases|CFC|CFC12",
                ]
            ],
        },
        "Emissions|Montreal Gases|CFC|CFC11": {
            "title": "CFC11 emissions",
            "comment": "total emissions of CFC11",
        },
        "Emissions|Montreal Gases|CFC|CFC113": {
            "title": "CFC113 emissions",
            "comment": "total emissions of CFC113",
        },
        "Emissions|Montreal Gases|CFC|CFC114": {
            "title": "CFC114 emissions",
            "comment": "total emissions of CFC114",
        },
        "Emissions|Montreal Gases|CFC|CFC115": {
            "title": "CFC115 emissions",
            "comment": "total emissions of CFC115",
        },
        "Emissions|Montreal Gases|CFC|CFC12": {
            "title": "CFC12 emissions",
            "comment": "total emissions of CFC12",
        },
        "Emissions|Montreal Gases|CH2Cl2": {
            "title": "CH2Cl2 emissions",
            "comment": "total emissions of CH2Cl2",
        },
        "Emissions|Montreal Gases|CH3Br": {
            "title": "CH3Br emissions",
            "comment": "total emissions of CH3Br",
        },
        "Emissions|Montreal Gases|CH3CCl3": {
            "title": "CH3CCl3 emissions",
            "comment": "total emissions of CH3CCl3",
        },
        "Emissions|Montreal Gases|CH3Cl": {
            "title": "CH3Cl emissions",
            "comment": "total emissions of CH3Cl",
        },
        "Emissions|Montreal Gases|CHCl3": {
            "title": "CHCl3 emissions",
            "comment": "total emissions of CHCl3",
        },
        "Emissions|Montreal Gases|Halon1202": {
            "title": "Halon-1202 emissions",
            "comment": "total emissions of Halon-1202",
        },
        "Emissions|Montreal Gases|Halon1211": {
            "title": "Halon-1211 emissions",
            "comment": "total emissions of Halon-1211",
        },
        "Emissions|Montreal Gases|Halon1301": {
            "title": "Halon-1301 emissions",
            "comment": "total emissions of Halon-1301",
        },
        "Emissions|Montreal Gases|Halon2402": {
            "title": "Halon-2402 emissions",
            "comment": "total emissions of Halon-2402",
        },
        "Emissions|Montreal Gases|HCFC141b": {
            "title": "HCFC141b emissions",
            "comment": "total emissions of HCFC141b",
        },
        "Emissions|Montreal Gases|HCFC142b": {
            "title": "HCFC22 emissions",
            "comment": "total emissions of HCFC22",
        },
        "Emissions|Montreal Gases|HCFC22": {
            "title": "HCFC22 emissions",
            "comment": "total emissions of HCFC22",
        },
        "Emissions|N2O": {
            "title": "Nitrogen emissions",
            "comment": "total nitrogen emissions",
            "children": [
                [
                    "Emissions|N2O|MAGICC AFOLU",
                    "Emissions|N2O|MAGICC Fossil and Industrial",
                    "Emissions|N2O|Other",
                ]
            ],
        },
        "Emissions|N2O|MAGICC AFOLU": {
            "title": "Nitrogen AFOLU emissions",
            "comment": "nitrogen emissions from agriculture, forestry and other land use (IPCC category 3), excluding any fossil-fuel based emissions in the Agricultural sector (hence not identical to WG3 AFOLU)",
        },
        "Emissions|N2O|MAGICC Fossil and Industrial": {
            "title": "Nitrogen fossil and industrial emissions",
            "comment": "nitrogen emissions from energy use on supply and demand side (IPCC category 1A, 1B), industrial processes (IPCC category 2), waste (IPCC category 4) and other (IPCC category 5)",
        },
        "Emissions|N2O|Other": {
            "title": "Nitrogen emissions from other sources",
            "comment": "nitrogen emissions from other sources (please provide a definition)",
        },
        "Emissions|NH3": {
            "title": "Ammonia emissions",
            "comment": "total ammonia emissions",
            "children": [
                [
                    "Emissions|NH3|MAGICC AFOLU",
                    "Emissions|NH3|MAGICC Fossil and Industrial",
                    "Emissions|NH3|Other",
                ]
            ],
        },
        "Emissions|NH3|MAGICC AFOLU": {
            "title": "Ammonia AFOLU emissions",
            "comment": "ammonia emissions from agriculture, forestry and other land use (IPCC category 3), excluding any fossil-fuel based emissions in the Agricultural sector (hence not identical to WG3 AFOLU)",
        },
        "Emissions|NH3|MAGICC Fossil and Industrial": {
            "title": "Ammonia fossil and industrial emissions",
            "comment": "ammonia emissions from energy use on supply and demand side (IPCC category 1A, 1B), industrial processes (IPCC category 2), waste (IPCC category 4) and other (IPCC category 5)",
        },
        "Emissions|NH3|Other": {
            "title": "Ammonia emissions from other sources",
            "comment": "ammonia emissions from other sources (please provide a definition)",
        },
        "Emissions|NOx": {
            "title": "Nitrous oxide emissions",
            "comment": "total nitrous oxide emissions",
            "children": [
                [
                    "Emissions|NOx|MAGICC AFOLU",
                    "Emissions|NOx|MAGICC Fossil and Industrial",
                    "Emissions|NOx|Other",
                ]
            ],
        },
        "Emissions|NOx|MAGICC AFOLU": {
            "title": "Nitrous oxide AFOLU emissions",
            "comment": "nitrous oxide emissions from agriculture, forestry and other land use (IPCC category 3), excluding any fossil-fuel based emissions in the Agricultural sector (hence not identical to WG3 AFOLU)",
        },
        "Emissions|NOx|MAGICC Fossil and Industrial": {
            "title": "Nitrous oxide fossil and industrial emissions",
            "comment": "nitrous oxide emissions from energy use on supply and demand side (IPCC category 1A, 1B), industrial processes (IPCC category 2), waste (IPCC category 4) and other (IPCC category 5)",
        },
        "Emissions|NOx|Other": {
            "title": "Nitrous oxide emissions from other sources",
            "comment": "nitrous oxide emissions from other sources (please provide a definition)",
        },
        "Emissions|OC": {
            "title": "Organic carbon emissions",
            "comment": "total organic carbon emissions",
            "children": [
                [
                    "Emissions|OC|MAGICC AFOLU",
                    "Emissions|OC|MAGICC Fossil and Industrial",
                    "Emissions|OC|Other",
                ]
            ],
        },
        "Emissions|OC|MAGICC AFOLU": {
            "title": "Organic carbon AFOLU emissions",
            "comment": "organic carbon emissions from agriculture, forestry and other land use (IPCC category 3), excluding any fossil-fuel based emissions in the Agricultural sector (hence not identical to WG3 AFOLU)",
        },
        "Emissions|OC|MAGICC Fossil and Industrial": {
            "title": "Organic carbon fossil and industrial emissions",
            "comment": "organic carbon emissions from energy use on supply and demand side (IPCC category 1A, 1B), industrial processes (IPCC category 2), waste (IPCC category 4) and other (IPCC category 5)",
        },
        "Emissions|OC|Other": {
            "title": "Organic carbon emissions from other sources",
            "comment": "organic carbon emissions from other sources (please provide a definition)",
        },
        "Emissions|Sulfur": {
            "title": "Sulfur emissions",
            "comment": "total sulfur (as a precursor for sulfates) emissions",
            "children": [
                [
                    "Emissions|Sulfur|MAGICC AFOLU",
                    "Emissions|Sulfur|MAGICC Fossil and Industrial",
                    "Emissions|Sulfur|Other",
                ]
            ],
        },
        "Emissions|Sulfur|MAGICC AFOLU": {
            "title": "Sulfur AFOLU emissions",
            "comment": "sulfur (as a precursor for sulfates) emissions from agriculture, forestry and other land use (IPCC category 3), excluding any fossil-fuel based emissions in the Agricultural sector (hence not identical to WG3 AFOLU)",
        },
        "Emissions|Sulfur|MAGICC Fossil and Industrial": {
            "title": "Sulfur fossil and industrial emissions",
            "comment": "sulfur (as a precursor for sulfates) emissions from energy use on supply and demand side (IPCC category 1A, 1B), industrial processes (IPCC category 2), waste (IPCC category 4) and other (IPCC category 5)",
        },
        "Emissions|Sulfur|Other": {
            "title": "Sulfur emissions from other sources",
            "comment": "sulfur (as a precursor for sulfates) emissions from other sources (please provide a definition)",
        },
        "Emissions|VOC": {
            "title": "(Non-methane) volatile organic compounds emissions",
            "comment": "total (non-methane) volatile organic compounds emissions",
            "children": [
                [
                    "Emissions|VOC|MAGICC AFOLU",
                    "Emissions|VOC|MAGICC Fossil and Industrial",
                    "Emissions|VOC|Other",
                ]
            ],
        },
        "Emissions|VOC|MAGICC AFOLU": {
            "title": "(Non-methane) volatile organic compounds AFOLU emissions",
            "comment": "(non-methane) volatile organic compounds emissions from agriculture, forestry and other land use (IPCC category 3), excluding any fossil-fuel based emissions in the Agricultural sector (hence not identical to WG3 AFOLU)",
        },
        "Emissions|VOC|MAGICC Fossil and Industrial": {
            "title": "(Non-methane) volatile organic compounds fossil and industrial emissions",
            "comment": "(non-methane) volatile organic compounds emissions from energy use on supply and demand side (IPCC category 1A, 1B), industrial processes (IPCC category 2), waste (IPCC category 4) and other (IPCC category 5)",
        },
        "Emissions|VOC|Other": {
            "title": "(Non-methane) volatile organic compounds emissions from other sources",
            "comment": "(non-methane) volatile organic compounds emissions from other sources (please provide a definition)",
        },
    },
}



"""Classes to represent and query categorical systems."""

import datetime
import functools
import importlib
import importlib.resources
import itertools
import pathlib
import pickle
import typing
from typing import TypeVar

import natsort
import networkx as nx
import pandas
import strictyaml as sy
from black import Mode, format_str
from ruamel.yaml import YAML

from . import data
from ._conversions import Conversion, ConversionSpec

# Categorization, or any subclass.
CategorizationT = TypeVar("CategorizationT", bound="Categorization")


class Category:
    """A single category."""

    _strictyaml_schema = sy.Map(
        {
            "title": sy.Str(),
            sy.Optional("comment"): sy.Str(),
            sy.Optional("alternative_codes"): sy.Seq(sy.Str()),
            sy.Optional("info"): sy.MapPattern(sy.Str(), sy.Any()),
        }
    )

    def __init__(
        self,
        codes: tuple[str, ...],
        categorization: "Categorization",
        title: str,
        comment: typing.Union[None, str] = None,
        info: typing.Union[None, dict] = None,
    ):
        self.codes = codes
        self.title = title
        self.comment = comment
        self.categorization = categorization
        if info is None:
            self.info = {}
        else:
            self.info = info
        self._hash = None

    @classmethod
    def from_spec(cls, code: str, spec: dict, categorization: "Categorization"):
        codes = [code]
        if "alternative_codes" in spec:
            codes += spec["alternative_codes"]
            del spec["alternative_codes"]
        return cls(
            codes=tuple(codes),
            categorization=categorization,
            title=spec["title"],
            comment=spec.get("comment"),
            info=spec.get("info"),
        )

    def to_spec(self) -> tuple[str, dict[str, typing.Union[str, dict, list]]]:
        """Turn this category into a specification ready to be written to a yaml file.

        Returns
        -------
        (code: str, spec: dict)
            Primary code and specification dict
        """
        code = self.codes[0]
        spec: dict[str, typing.Union[str, dict, list[str]]] = {"title": self.title}
        if self.comment is not None:
            spec["comment"] = self.comment
        if len(self.codes) > 1:
            spec["alternative_codes"] = list(self.codes[1:])
        if self.info:
            spec["info"] = self.info
        return code, spec

    def __str__(self) -> str:
        return f"{self.codes[0]} {self.title}"

    def __eq__(self, other: object):
        if not isinstance(other, Category):
            return NotImplemented
        return any(x in other.codes for x in self.codes) and (
            self.categorization is other.categorization
            or self.categorization.name.startswith(f"{other.categorization.name}_")
            or other.categorization.name.startswith(f"{self.categorization.name}_")
            or self.categorization.name == other.categorization.name
        )

    def __repr__(self) -> str:
        return f"<{self.categorization.name}: {self.codes[0]!r}>"

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self.categorization.name + self.codes[0])
        return self._hash

    def __lt__(self, other):
        s = natsort.natsorted((self.codes[0], other.codes[0]))
        return s[0] == self.codes[0] and self != other


class HierarchicalCategory(Category):
    """A single category from a HierarchicalCategorization."""

    _strictyaml_schema = sy.Map(
        {
            "title": sy.Str(),
            sy.Optional("comment"): sy.Str(),
            sy.Optional("alternative_codes"): sy.Seq(sy.Str()),
            sy.Optional("info"): sy.MapPattern(sy.Str(), sy.Any()),
            sy.Optional("children"): sy.Seq(sy.Seq(sy.Str())),
        }
    )

    def __init__(
        self,
        codes: tuple[str],
        categorization: "HierarchicalCategorization",
        title: str,
        comment: typing.Union[None, str] = None,
        info: typing.Union[None, dict] = None,
    ):
        Category.__init__(self, codes, categorization, title, comment, info)
        self.categorization = categorization

    def to_spec(self) -> tuple[str, dict[str, typing.Union[str, dict, list]]]:
        """Turn this category into a specification ready to be written to a yaml file.

        Returns
        -------
        (code: str, spec: dict)
            Primary code and specification dict
        """
        code, spec = Category.to_spec(self)
        children = [
            list(sorted(c.codes[0] for c in child_set)) for child_set in self.children
        ]
        if children:
            spec["children"] = children
        return code, spec

    @property
    def children(self) -> list[set["HierarchicalCategory"]]:
        """The sets of subcategories comprising this category.

        The first set is canonical, the other sets are alternative.
        Only the canonical sets are used to calculate the level of a category."""
        return self.categorization.children(self)

    @property
    def parents(self) -> set["HierarchicalCategory"]:
        """The super-categories where this category is a member of any set of children.

        Note that all possible parents are returned, not "canonical" parents.
        """
        return self.categorization.parents(self)

    @property
    def ancestors(self) -> set["HierarchicalCategory"]:
        """The super-categories where this category or any of its parents is a member
        of any set of children, transitively.

        Note that all possible ancestors are returned, not only "canonical" ones.
        """
        return self.categorization.ancestors(self)

    @property
    def descendants(self) -> set["HierarchicalCategory"]:
        """The sets of subcategories comprising this category directly or indirectly.

        Note that all possible descendants are returned, not only "canonical" ones."""
        return self.categorization.descendants(self)

    @property
    def is_leaf(self) -> bool:
        """Is this category a leaf category, i.e. without children?"""
        return not any(self.children)

    @property
    def leaf_children(self) -> list[set["HierarchicalCategory"]]:
        """The sets of subcategories which are descendants of this category and do not
        have children themselves.

        Sets of children are chased separately, so each set of leaf children is
        self-sufficient to reconstruct this category (if the categorization allows
        reconstructing categories from their children, i.e. if total_sum is set)."""
        ret = []
        for children in self.children:
            n = []
            for child in children:
                if child.is_leaf:
                    n.append([{child}])
                else:
                    n.append(child.leaf_children)
            ret += [set(itertools.chain(*x)) for x in itertools.product(*n)]
        return ret

    @property
    def level(self) -> int:
        """The level of the category.

        The canonical top-level category has level 1 and its children have level 2 etc.

        To calculate the level, only the first ("canonical") set of children is
        considered for intermediate categories.
        """
        return self.categorization.level(self)


class Categorization:
    """A single categorization system.

    A categorization system comprises a set of categories, and their relationships as
    well as metadata describing the categorization system itself.

    Use the categorization object like a dictionary, where codes can be translated
    to their meaning using ``cat[code]`` and all codes are available using
    ``cat.keys()``. Metadata about the categorization is provided in attributes.
    If `pandas` is available, you can access a `pandas.DataFrame` with all
    category codes, and their meanings at ``cat.df``.

    Attributes
    ----------
    name : str
        The unique name/code
    references : str
        Citable reference(s)
    title : str
        A short, descriptive title for humans
    comment : str
        Notes and explanations for humans
    institution : str
        Where the categorization originates
    last_update : datetime.date
        The date of the last change
    version : str, optional
        The version of the Categorization, if there are multiple versions
    hierarchical : bool
        True if descendants and ancestors are defined
    """

    hierarchical: bool = False

    _strictyaml_schema = sy.Map(
        {
            "name": sy.Str(),
            "title": sy.Str(),
            "comment": sy.Str(),
            "references": sy.Str(),
            "institution": sy.Str(),
            "last_update": sy.Str(),
            "hierarchical": sy.Bool(),
            sy.Optional("version"): sy.Str(),
            "categories": sy.MapPattern(sy.Str(), Category._strictyaml_schema),
        }
    )

    def _add_categories(self, categories: dict[str, dict]):
        for code, spec in categories.items():
            cat = Category.from_spec(code=code, spec=spec, categorization=self)

            self._primary_code_map[code] = cat
            for icode in cat.codes:
                self._all_codes_map[icode] = cat

    def __init__(
        self,
        *,
        categories: dict[str, dict],
        name: str,
        title: str,
        comment: str,
        references: str,
        institution: str,
        last_update: datetime.date,
        version: typing.Union[None, str] = None,
    ):
        self._primary_code_map: dict[str, Category] = {}
        self._all_codes_map: dict[str, Category] = {}
        self.name = name
        self.references = references
        self.title = title
        self.comment = comment
        self.institution = institution
        self.last_update = last_update
        self.version = version

        self._add_categories(categories)

        # is filled in __init__.py to contain all categorizations
        self._cats: dict[str, Categorization] = {}

    def __hash__(self):
        return hash(self.name)

    @classmethod
    def from_yaml(
        cls: type[CategorizationT],
        filepath: typing.Union[str, pathlib.Path, typing.TextIO],
    ) -> CategorizationT:
        """Read Categorization from a StrictYaml file."""
        try:
            yaml = sy.load(filepath.read(), schema=cls._strictyaml_schema)
        except AttributeError:
            with open(filepath) as fd:
                yaml = sy.load(fd.read(), schema=cls._strictyaml_schema)
        return cls.from_spec(yaml.data)

    @classmethod
    def from_spec(
        cls: type[CategorizationT], spec: dict[str, typing.Any]
    ) -> CategorizationT:
        """Create Categorization from a Dictionary specification."""
        if spec["hierarchical"] != cls.hierarchical:
            raise ValueError(
                "Specification is for a hierarchical categorization, use"
                "HierarchicalCategorization.from_spec."
            )
        last_update = datetime.date.fromisoformat(spec["last_update"])
        return cls(
            categories=spec["categories"],
            name=spec["name"],
            title=spec["title"],
            comment=spec["comment"],
            references=spec["references"],
            institution=spec["institution"],
            last_update=last_update,
            version=spec.get("version"),
        )

    @staticmethod
    def from_pickle(
        filepath: typing.Union[str, pathlib.Path, typing.IO[bytes]],
    ) -> CategorizationT:
        """De-serialize Categorization from a file written by to_pickle.

        Note that this uses the pickle module, which executes arbitrary code in the
        provided file. Only load from pickle files that you trust."""
        return from_pickle(filepath)

    @staticmethod
    def from_python(
        filepath: typing.Union[str, pathlib.Path, typing.IO[bytes]],
    ) -> CategorizationT:
        """De-serialize Categorization from a file written by to_python.

        Note that this executes the python cache file. Only load from python cache files
        you trust."""
        return from_python(filepath)

    def to_spec(self) -> dict[str, typing.Any]:
        """Turn this categorization into a specification dictionary ready to be written
        to a yaml file.

        Returns
        -------
        spec: dict
            Specification dictionary understood by `from_spec`.
        """
        spec = {
            "name": self.name,
            "title": self.title,
            "comment": self.comment,
            "references": self.references,
            "institution": self.institution,
            "hierarchical": self.hierarchical,
            "last_update": self.last_update.isoformat(),
        }
        if self.version is not None:
            spec["version"] = self.version
        categories = {}
        for cat in self.values():
            code, cat_spec = cat.to_spec()
            categories[code] = cat_spec
        spec["categories"] = categories

        return spec

    def to_yaml(self, filepath: typing.Union[str, pathlib.Path]) -> None:
        """Write to a YAML file."""
        spec = self.to_spec()
        yaml = YAML()
        yaml.default_flow_style = False
        with open(filepath, "w") as fd:
            yaml.dump(spec, fd)

    def to_python(self, filepath: typing.Union[str, pathlib.Path]) -> None:
        """Write spec to a Python file."""
        spec = self.to_spec()
        comment = (
            "# Do not edit this file. It was auto-generated from the\n"
            "# corresponding YAML file.\n"
        )
        with open(filepath, "w") as f:
            f.write(comment)
            f.write(f"spec = {format_str(repr(spec), mode=Mode())}")

    def to_pickle(self, filepath: typing.Union[str, pathlib.Path]) -> None:
        """Serialize to a file using python's pickle."""
        spec = self.to_spec()
        with open(filepath, "wb") as fd:
            pickle.dump(spec, fd, protocol=4)

    def keys(self) -> typing.KeysView[str]:
        """Iterate over the codes for all categories."""
        return self._primary_code_map.keys()

    def values(self) -> typing.ValuesView[Category]:
        """Iterate over the categories."""
        return self._primary_code_map.values()

    def items(self) -> typing.ItemsView[str, Category]:
        """Iterate over (primary code, category) pairs."""
        return self._primary_code_map.items()

    def all_keys(self) -> typing.KeysView[str]:
        """Iterate over all codes for all categories."""
        return self._all_codes_map.keys()

    def __iter__(self) -> typing.Iterable[str]:
        return iter(self._primary_code_map)

    def __getitem__(self, code: str) -> Category:
        """Get the category for a code."""
        return self._all_codes_map[code]

    def __contains__(self, code: str) -> bool:
        """Can the code be mapped to a category?"""
        return code in self._all_codes_map

    def __len__(self) -> int:
        return len(self._primary_code_map)

    def __repr__(self) -> str:
        return (
            f"<Categorization {self.name} {self.title!r} with {len(self)} categories>"
        )

    def __str__(self) -> str:
        return self.name

    @property
    def df(self) -> "pandas.DataFrame":
        """All category codes as a pandas dataframe."""
        titles = []
        comments = []
        alternative_codes = []
        for cat in self.values():
            titles.append(cat.title)
            comments.append(cat.comment)
            alternative_codes.append(cat.codes[1:])
        return pandas.DataFrame(
            index=list(self.keys()),
            data={
                "title": titles,
                "comment": comments,
                "alternative_codes": alternative_codes,
            },
        )

    def _extend_prepare(
        self,
        *,
        categories: typing.Union[None, dict[str, dict]] = None,
        alternative_codes: typing.Union[None, dict[str, str]] = None,
        name: str,
        title: typing.Union[None, str] = None,
        comment: typing.Union[None, str] = None,
        last_update: typing.Union[None, datetime.date] = None,
    ) -> dict[str, typing.Any]:
        spec = self.to_spec()

        spec["name"] = f"{self.name}_{name}"
        spec["references"] = ""
        spec["institution"] = ""

        if title is None:
            spec["title"] = f"{self.title} + {name}"
        else:
            spec["title"] = self.title + title

        if comment is None:
            spec["comment"] = f"{self.comment} extended by {name}"
        else:
            spec["comment"] = self.comment + comment

        if last_update is None:
            spec["last_update"] = datetime.date.today().isoformat()
        else:
            spec["last_update"] = last_update.isoformat()

        if categories is not None:
            spec["categories"].update(categories)

        if alternative_codes is not None:
            for alias, primary in alternative_codes.items():
                if "alternative_codes" not in spec["categories"][primary]:
                    spec["categories"][primary]["alternative_codes"] = []

                spec["categories"][primary]["alternative_codes"].append(alias)

        return spec

    def extend(
        self: CategorizationT,
        *,
        categories: typing.Union[None, dict[str, dict]] = None,
        alternative_codes: typing.Union[None, dict[str, str]] = None,
        name: str,
        title: typing.Union[None, str] = None,
        comment: typing.Union[None, str] = None,
        last_update: typing.Union[None, datetime.date] = None,
    ) -> CategorizationT:
        """Extend the categorization with additional categories, yielding a new
        categorization.

        Metadata: the ``name``, ``title``, ``comment``, and ``last_update`` are updated
        automatically (see below), the ``institution`` and ``references`` are deleted
        and the values for ``version`` and ``hierarchical`` are kept.
        You can set more accurate metadata (for example, your institution) on the
        returned object if needed.

        Parameters
        ----------
        categories: dict, optional
           Map of new category codes to their specification. The specification is a
           dictionary with the keys "title", optionally "comment", and optionally
           "alternative_codes".
        alternative_codes: dict, optional
           Map of new alternative codes. A dictionary with the new alternative code
           as key and existing code as value.
        name : str
           The name of your extension. The returned Categorization will have a name
           of "{old_name}_{name}", indicating that it is an extension of the underlying
           Categorization.
        title : str, optional
           A string to add to the original title. If not provided, " + {name}" will be
           used.
        comment : str, optional
           A string to add to the original comment. If not provided,
           " extended by {name}" will be used.
        last_update : datetime.date, optional
           The date of the last update to this extension. Today will be used if not
           provided.

        Returns
        -------
        Extended categorization : Categorization
        """
        spec = self._extend_prepare(
            name=name,
            categories=categories,
            title=title,
            comment=comment,
            last_update=last_update,
            alternative_codes=alternative_codes,
        )

        return Categorization.from_spec(spec)

    def __eq__(self, other):
        if not isinstance(other, Categorization):
            return False
        if self.name != other.name:
            return False
        return self._primary_code_map == other._primary_code_map

    def conversion_to(self, other: typing.Union["Categorization", str]) -> Conversion:
        """Get conversion to other categorization.

        If conversion rules for this conversion are not included, raises
        NotImplementedError."""
        if isinstance(other, str):
            other_name = other
        else:
            other_name = other.name

        data_files = importlib.resources.files(data)
        forward_file = data_files / f"conversion.{self.name}.{other_name}.csv"
        if forward_file.is_file():
            return ConversionSpec.from_csv(forward_file.open()).hydrate(cats=self._cats)
        reverse_file = data_files / f"conversion.{other_name}.{self.name}.csv"
        if reverse_file.is_file():
            return (
                ConversionSpec.from_csv(reverse_file.open())
                .hydrate(cats=self._cats)
                .reversed()
            )

        raise NotImplementedError(
            f"Conversion between {self.name} and {other_name} not yet included."
        )


class HierarchicalCategorization(Categorization):
    """In a hierarchical categorization, descendants and ancestors (parents and
    children) are defined for each category.

    Attributes
    ----------
    total_sum : bool
        If the sum of the values of children equals the value of the parent for
        extensive quantities. For example, a Categorization containing the Countries in
        the EU and the EU could set `total_sum = True`, because the emissions of all
        parts of the EU must equal the emissions of the EU. On the contrary, a
        categorization of Industries with categories `Power:Fossil Fuels` and
        `Power:Gas` which are both children of `Power` must set `total_sum = False`
        to avoid double counting of fossil gas.
    canonical_top_level_category : HierarchicalCategory
        The level of a category is calculated with respect to the canonical top level
        category. Commonly, this will be the world total or a similar category. If the
        canonical top level category is not set (i.e. is ``None``), levels are not
        defined for categories.
    """

    hierarchical = True

    _strictyaml_schema = sy.Map(
        {
            "name": sy.Str(),
            "title": sy.Str(),
            "comment": sy.Str(),
            "references": sy.Str(),
            "institution": sy.Str(),
            "last_update": sy.Str(),
            "hierarchical": sy.Bool(),
            sy.Optional("version"): sy.Str(),
            "total_sum": sy.Bool(),
            sy.Optional("canonical_top_level_category"): sy.Str(),
            "categories": sy.MapPattern(
                sy.Str(), HierarchicalCategory._strictyaml_schema
            ),
        }
    )

    def _add_categories(self, categories: dict[str, dict]):
        for code, spec in categories.items():
            cat = HierarchicalCategory.from_spec(
                code=code, spec=spec, categorization=self
            )

            self._primary_code_map[code] = cat
            self._graph.add_node(cat)
            for icode in cat.codes:
                self._all_codes_map[icode] = cat

        for code, spec in categories.items():
            if "children" in spec:
                parent = self._all_codes_map[code]
                for i, child_set in enumerate(spec["children"]):
                    for child_code in child_set:
                        self._graph.add_edge(
                            parent, self._all_codes_map[child_code], set=i
                        )

    def __init__(
        self,
        *,
        categories: dict[str, dict],
        name: str,
        title: str,
        comment: str,
        references: str,
        institution: str,
        last_update: datetime.date,
        version: typing.Union[None, str] = None,
        total_sum: bool,
        canonical_top_level_category: typing.Union[None, str] = None,
    ):
        self._graph = nx.MultiDiGraph()
        Categorization.__init__(
            self,
            categories=categories,
            name=name,
            title=title,
            comment=comment,
            references=references,
            institution=institution,
            last_update=last_update,
            version=version,
        )
        self.total_sum = total_sum
        if canonical_top_level_category is None:
            self.canonical_top_level_category: typing.Union[
                None, HierarchicalCategory
            ] = None
        else:
            self.canonical_top_level_category = self._all_codes_map[
                canonical_top_level_category
            ]

    def __getitem__(self, code: str) -> HierarchicalCategory:
        """Get the category for a code."""
        return self._all_codes_map[code]

    def values(self) -> typing.ValuesView[HierarchicalCategory]:
        """Iterate over the categories."""
        return self._primary_code_map.values()

    def items(self) -> typing.ItemsView[str, HierarchicalCategory]:
        """Iterate over (primary code, category) pairs."""
        return self._primary_code_map.items()

    @classmethod
    def from_spec(
        cls: type[CategorizationT], spec: dict[str, typing.Any]
    ) -> CategorizationT:
        """Create Categorization from a Dictionary specification."""
        if spec["hierarchical"] != cls.hierarchical:
            raise ValueError(
                "Specification is for a non-hierarchical categorization, use"
                "Categorization.from_spec."
            )
        last_update = datetime.date.fromisoformat(spec["last_update"])
        return cls(
            categories=spec["categories"],
            name=spec["name"],
            title=spec["title"],
            comment=spec["comment"],
            references=spec["references"],
            institution=spec["institution"],
            last_update=last_update,
            version=spec.get("version"),
            total_sum=spec["total_sum"],
            canonical_top_level_category=spec.get("canonical_top_level_category"),
        )

    def to_spec(self) -> dict[str, typing.Any]:
        """Turn this categorization into a specification dictionary ready to be written
        to a yaml file.

        Returns
        -------
        spec: dict
            Specification dictionary understood by `from_spec`.
        """
        # we can't call Categorization.to_spec here because we need to control ordering
        # in the returned dict so that we get nicely ordered yaml files.
        spec = {
            "name": self.name,
            "title": self.title,
            "comment": self.comment,
            "references": self.references,
            "institution": self.institution,
            "hierarchical": self.hierarchical,
            "last_update": self.last_update.isoformat(),
        }
        if self.version is not None:
            spec["version"] = self.version
        spec["total_sum"] = self.total_sum
        if self.canonical_top_level_category is not None:
            spec["canonical_top_level_category"] = (
                self.canonical_top_level_category.codes[0]
            )

        spec["categories"] = {}
        for cat in self.values():
            code, cat_spec = cat.to_spec()
            spec["categories"][code] = cat_spec

        return spec

    @functools.cached_property
    def _canonical_subgraph(self) -> nx.DiGraph:
        return nx.DiGraph(
            self._graph.edge_subgraph(
                (u, v, 0) for (u, v, s) in self._graph.edges(data="set") if s == 0
            )
        )

    def _show_subtree_children(
        self,
        children: typing.Iterable[HierarchicalCategory],
        format_func: typing.Callable,
        prefix: str,
        maxdepth: typing.Union[None, int],
    ) -> str:
        children_sorted = natsort.natsorted(children, key=format_func)
        r = "".join(
            self._show_subtree(
                node=child,
                prefix=f"{prefix}│",
                format_func=format_func,
                maxdepth=maxdepth,
            )
            for child in children_sorted[:-1]
        )
        # Last child needs to be called slightly differently
        r += self._show_subtree(
            node=children_sorted[-1],
            prefix=f"{prefix} ",
            last=True,
            format_func=format_func,
            maxdepth=maxdepth,
        )
        return r

    @staticmethod
    def _render_node(
        node: HierarchicalCategory,
        last: bool,
        prefix: str,
        format_func: typing.Callable[[HierarchicalCategory], str],
    ):
        formatted = format_func(node)
        if prefix:
            if last:
                return f"{prefix[:-1]}╰{formatted}\n"
            else:
                return f"{prefix[:-1]}├{formatted}\n"
        else:
            return f"{formatted}\n"

    def _show_subtree(
        self,
        *,
        node: HierarchicalCategory,
        prefix="",
        last=False,
        format_func: typing.Callable[[HierarchicalCategory], str] = str,
        maxdepth: typing.Union[None, int],
    ) -> str:
        """Recursively-called function to show a subtree starting at the given node."""

        r = self._render_node(node, last=last, prefix=prefix, format_func=format_func)

        if maxdepth is not None:
            maxdepth -= 1
            if maxdepth == 0:  # maxdepth reached, nothing more to do
                return r

        child_sets = node.children
        if len(child_sets) == 1:
            children = child_sets[0]
            if children:
                r += self._show_subtree_children(
                    children=children,
                    format_func=format_func,
                    maxdepth=maxdepth,
                    prefix=prefix,
                )
        elif len(child_sets) > 1:
            prefix += "║"
            i = 1
            for children in child_sets:
                if children:
                    if i == 1:
                        r += (
                            f"{prefix[:-1]}╠╤══ ('{format_func(node)}'s children,"
                            f" option 1)\n"
                        )
                    else:
                        r += (
                            f"{prefix[:-1]}╠╕ ('{format_func(node)}'s children,"
                            f" option {i})\n"
                        )

                    r += self._show_subtree_children(
                        children=children,
                        format_func=format_func,
                        maxdepth=maxdepth,
                        prefix=prefix,
                    )
                    i += 1

            r += f"{prefix[:-1]}╚═══\n"

        return r

    def show_as_tree(
        self,
        *,
        format_func: typing.Callable[[HierarchicalCategory], str] = str,
        maxdepth: typing.Union[None, int] = None,
        root: typing.Union[None, HierarchicalCategory, str] = None,
    ) -> str:
        """Format the hierarchy as a tree.

        Starting from the given root, or - if no root is given - the top-level
        categories (i.e. categories without parents), the tree of categories that are
        transitive children of the root is show, with children connected to their
        parents using lines. If a parent category has one set of children, the children
        are connected to each other and the parent with a simple line. If a parent
        category has multiple sets of children, the sets are connected to parent with
        double lines and the children in a set are connected to each other with simple
        lines.

        Parameters
        ----------
        format_func: callable, optional
            Function to call to format categories for display. Each category is
            formatted for display using format_func(category), so format_func should
            return a string without line breaks, otherwise the tree will look weird.
            By default, str() is used, so that the first code and the title of the
            category are used.
        maxdepth: int, optional
            Maximum depth to show in the tree. By default, goes to arbitrary depth.
        root: HierarchicalCategory or str, optional
            HierarchicalCategory object or code to use as the top-most category.
            If not given, the whole tree is shown, starting from all categories without
            parents.

        Returns
        -------
        tree_str: str
            Representation of the hierarchy as formatted string. print() it for optimal
            viewing.
        """
        if root is None:
            top_level_nodes = (node for node in self.values() if not node.parents)
        else:
            if not isinstance(root, HierarchicalCategory):
                root = self[root]
            top_level_nodes = [root]
        return "\n".join(
            (
                self._show_subtree(
                    node=top_level_node, format_func=format_func, maxdepth=maxdepth
                )
            )
            for top_level_node in top_level_nodes
        )

    def extend(
        self,
        *,
        categories: typing.Union[None, dict[str, dict]] = None,
        alternative_codes: typing.Union[None, dict[str, str]] = None,
        children: typing.Union[None, list[tuple]] = None,
        name: str,
        title: typing.Union[None, str] = None,
        comment: typing.Union[None, str] = None,
        last_update: typing.Union[None, datetime.date] = None,
    ) -> "HierarchicalCategorization":
        """Extend the categorization with additional categories and relationships,
        yielding a new categorization.

        Metadata: the ``name``, ``title``, ``comment``, and ``last_update`` are updated
        automatically (see below), the ``institution`` and ``references`` are deleted
        and the values for ``version``, ``hierarchical``, ``total_sum``, and
        ``canonical_top_level_category`` are kept.
        You can set more accurate metadata (for example, your institution) on the
        returned object if needed.

        Parameters
        ----------
        categories: dict, optional
           Map of new category codes to their specification. The specification is a
           dictionary with the keys "title", optionally "comment", and optionally
           "alternative_codes".
        alternative_codes: dict, optional
           Map of new alternative codes. A dictionary with the new alternative code
           as key and existing code as value.
        children: list, optional
           List of ``(parent, (child1, child2, …))`` pairs. The given relationships will
           be inserted in the extended categorization.
        name : str
           The name of your extension. The returned Categorization will have a name
           of "{old_name}_{name}", indicating that it is an extension of the underlying
           Categorization.
        title : str, optional
           A string to add to the original title. If not provided, " + {name}" will be
           used.
        comment : str, optional
           A string to add to the original comment. If not provided,
           " extended by {name}" will be used.
        last_update : datetime.date, optional
           The date of the last update to this extension. Today will be used if not
           provided.

        Returns
        -------
        Extended categorization : HierarchicalCategorization
        """
        spec = self._extend_prepare(
            name=name,
            categories=categories,
            title=title,
            comment=comment,
            last_update=last_update,
            alternative_codes=alternative_codes,
        )

        if children is not None:
            for parent, child_set in children:
                if "children" not in spec["categories"][parent]:
                    spec["categories"][parent]["children"] = []
                spec["categories"][parent]["children"].append(child_set)

        return HierarchicalCategorization.from_spec(spec)

    @property
    def df(self) -> "pandas.DataFrame":
        """All category codes as a pandas dataframe."""
        titles = []
        comments = []
        alternative_codes = []
        children = []
        for cat in self.values():
            titles.append(cat.title)
            comments.append(cat.comment)
            alternative_codes.append(cat.codes[1:])
            children.append(
                tuple(tuple(sorted(c.codes[0] for c in cs)) for cs in cat.children)
            )
        return pandas.DataFrame(
            index=self.keys(),
            data={
                "title": titles,
                "comment": comments,
                "alternative_codes": alternative_codes,
                "children": children,
            },
        )

    def level(self, cat: typing.Union[str, HierarchicalCategory]) -> int:
        """The level of the given category.

        The canonical top-level category has level 1 and its children have level 2 etc.

        To calculate the level, first only the first ("canonical") set of children is
        considered. Only if no path from the canonical top-level category to the
        given category can be found all other sets of children are considered to
        calculate the level.
        """
        if not isinstance(cat, HierarchicalCategory):
            return self.level(self[cat])
        if not isinstance(self.canonical_top_level_category, HierarchicalCategory):
            raise ValueError(
                "Can not calculate the level without a canonical_top_level_category."
            )

        # first use the canonical subgraph for shortest paths
        csg = self._canonical_subgraph
        try:
            sp = nx.shortest_path_length(csg, self.canonical_top_level_category, cat)
        except (nx.NetworkXNoPath, nx.NodeNotFound):
            try:
                sp = nx.shortest_path_length(
                    self._graph, self.canonical_top_level_category, cat
                )
            except (nx.NetworkXNoPath, nx.NodeNotFound):
                raise ValueError(
                    f"{cat.codes[0]!r} is not a transitive child of the "
                    f"canonical top level "
                    f"{self.canonical_top_level_category.codes[0]!r}."
                ) from None

        return sp + 1

    def parents(
        self, cat: typing.Union[str, HierarchicalCategory]
    ) -> set[HierarchicalCategory]:
        """The direct parents of the given category."""
        if not isinstance(cat, HierarchicalCategory):
            return self.parents(self._all_codes_map[cat])

        return set(self._graph.predecessors(cat))

    def ancestors(
        self, cat: typing.Union[str, HierarchicalCategory]
    ) -> set[HierarchicalCategory]:
        """All ancestors of the given category, i.e. the direct parents and their
        parents, etc."""
        if not isinstance(cat, HierarchicalCategory):
            return self.ancestors(self._all_codes_map[cat])

        return set(nx.ancestors(self._graph, cat))

    def children(
        self, cat: typing.Union[str, HierarchicalCategory]
    ) -> list[set[HierarchicalCategory]]:
        """The list of sets of direct children of the given category."""
        if not isinstance(cat, HierarchicalCategory):
            return self.children(self._all_codes_map[cat])

        children_dict = {}
        for _, child, setno in self._graph.edges(cat, "set"):
            if setno not in children_dict:
                children_dict[setno] = []
            children_dict[setno].append(child)

        return [set(children_dict[x]) for x in sorted(children_dict.keys())]

    def descendants(
        self, cat: typing.Union[str, HierarchicalCategory]
    ) -> set[HierarchicalCategory]:
        """All descendants of the given category, i.e. the direct children and their
        children, etc."""
        if not isinstance(cat, HierarchicalCategory):
            return self.descendants(self._all_codes_map[cat])

        return set(nx.descendants(self._graph, cat))

    def is_leaf(self, cat: typing.Union[str, HierarchicalCategory]) -> bool:
        """Is the category a leaf category, i.e. without children?"""
        if not isinstance(cat, HierarchicalCategory):
            return self.is_leaf(self._all_codes_map[cat])

        return cat.is_leaf

    def leaf_children(
        self, cat: typing.Union[str, HierarchicalCategory]
    ) -> list[set[HierarchicalCategory]]:
        """The sets of subcategories which are descendants of the category and do not
        have children themselves.

        Sets of children are chased separately, so each set of leaf children is
        self-sufficient to reconstruct this category (if the categorization allows
        reconstructing categories from their children, i.e. if total_sum is set)."""
        if not isinstance(cat, HierarchicalCategory):
            return self.leaf_children(self._all_codes_map[cat])

        return cat.leaf_children


def from_pickle(
    filepath: typing.Union[str, pathlib.Path, typing.IO[bytes]],
) -> typing.Union[Categorization, HierarchicalCategorization]:
    """De-serialize Categorization or HierarchicalCategorization from a file written by
    to_pickle.

    Note that this uses the pickle module, which executes arbitrary code in the
    provided file. Only load from pickle files that you trust."""
    try:
        spec = pickle.load(filepath)
    except TypeError:
        with open(filepath, "rb") as fd:
            spec = pickle.load(fd)

    return from_spec(spec)


def from_python(
    filepath: typing.Union[str, pathlib.Path, typing.IO[bytes]],
) -> CategorizationT:
    """Read Categorization or HierarchicalCategorization from a python cache file.

    Note that this executes the python cache file. Only load from python cache files
    you trust."""
    try:
        python_code = filepath.read()
        filepath.seek(0)
    except AttributeError:
        python_code = pathlib.Path(filepath).read_text()
    variables = {}
    exec(python_code, variables)
    spec = variables["spec"]
    if spec["hierarchical"]:
        cls = HierarchicalCategorization
    else:
        cls = Categorization
    return cls.from_spec(spec)


def from_spec(spec: dict[str, typing.Any]) -> CategorizationT:
    """Create Categorization or HierarchicalCategorization from a dict specification."""
    if spec["hierarchical"]:
        return HierarchicalCategorization.from_spec(spec)
    else:
        return Categorization.from_spec(spec)


def from_yaml(
    filepath: typing.Union[str, pathlib.Path, typing.TextIO],
) -> CategorizationT:
    """Read Categorization or HierarchicalCategorization from a StrictYaml file."""
    try:
        yaml = sy.load(filepath.read())
        filepath.seek(0)
    except AttributeError:
        with open(filepath) as fd:
            yaml = sy.load(fd.read())
    hier = yaml.data["hierarchical"]
    if hier in ("yes", "true", "True"):
        cls = HierarchicalCategorization
    elif hier in ("no", "false", "False"):
        cls = Categorization
    else:
        raise ValueError(
            f"'hierarchical' must be 'yes', 'true', 'True', 'no', 'false' or 'False',"
            f" not {hier!r}."
        )
    return cls.from_yaml(filepath)



"""Classes to represent conversions between categorizations."""

import csv
import dataclasses
import datetime
import pathlib
import typing
from typing import TYPE_CHECKING

import immutables
import pyparsing
import strictyaml as sy

if TYPE_CHECKING:
    from ._categories import Categorization, Category, HierarchicalCategory


@dataclasses.dataclass(frozen=True)
class ConversionRuleSpec:
    """Specification of a rule to convert between categories from two different
     categorizations.

    Supports one-to-one relationships, one-to-many relationships in both directions and
    many-to-many relationships. For each category, a factor is given which can also be
    negative to model relationships like A = B - C.

    Using auxiliary_categories, a rule can be restricted to specific auxiliary
    categories only.

    Attributes
    ----------
    factors_categories_a : dict mapping codes to factors
        Map of category codes from the first categorization to factors. For a simple
        addition, use factor 1, to subtract the category, use factor -1.
    factors_categories_b : dict mapping codes to factors
        Map of category codes from the second categorization to factors. For a simple
        addition, use factor 1, to subtract the category, use factor -1.
    auxiliary_categories : dict[str, set[str]]
        Map of auxiliary categorization names to sets of auxiliary category codes. Not
        all auxiliary categorizations need to be specified, and if an auxiliary
        categorization is not specified (or an empty set of category codes is given),
        the validity of the rule is not restricted.
        If an auxiliary categorization is specified and category codes are given, the
        rule is only valid for the given category codes. If multiple auxiliary
        categorizations are given, the rule is only valid if all auxiliary
        categorizations match.
    comment : str
        A human-readable comment explaining the rule or adding additional information.
    """

    factors_categories_a: dict[str, int]
    factors_categories_b: dict[str, int]
    auxiliary_categories: dict[str, set[str]]
    comment: str = ""
    csv_line_number: typing.Optional[int] = None
    csv_original_text: typing.Optional[str] = None

    def hydrate(
        self,
        categorization_a: "Categorization",
        categorization_b: "Categorization",
        cats: dict[str, "Categorization"],
    ) -> "ConversionRule":
        """Convert this specification into a ConversionRule object with full
        functionality."""

        auxiliary_categories_hydrated = {}
        for aux_categorization_name, categories in self.auxiliary_categories.items():
            aux_categorization = cats[aux_categorization_name]
            auxiliary_categories_hydrated[aux_categorization] = (
                self._hydrate_handle_errors(categories, aux_categorization)
            )

        return ConversionRule(
            factors_categories_a=self._hydrate_handle_errors(
                self.factors_categories_a, categorization_a
            ),
            factors_categories_b=self._hydrate_handle_errors(
                self.factors_categories_b, categorization_b
            ),
            auxiliary_categories=auxiliary_categories_hydrated,
            comment=self.comment,
            csv_line_number=self.csv_line_number,
            csv_original_text=self.csv_original_text,
        )

    @typing.overload
    def _hydrate_handle_errors(
        self, to_hydrate: dict[str, int], categorization: "Categorization"
    ) -> dict["Category", int]: ...

    @typing.overload
    def _hydrate_handle_errors(
        self, to_hydrate: set[str], categorization: "Categorization"
    ) -> set["Category"]: ...

    def _hydrate_handle_errors(
        self,
        to_hydrate: typing.Union[dict[str, int], set[str]],
        categorization: "Categorization",
    ) -> typing.Union[dict["Category", int], set["Category"]]:
        """Hydrate a dict/set while nicely handling errors."""
        try:
            if isinstance(to_hydrate, dict):
                return {
                    categorization[code]: factor for code, factor in to_hydrate.items()
                }
            else:
                return {categorization[code] for code in to_hydrate}
        except KeyError as err:
            code = err.args[0]
            raise ValueError(
                f"Error in line {self.csv_line_number}: {code!r} not in"
                f" {categorization}."
            ) from None

    # Parsing rules for simple formulas from str
    # Supported operators at the moment are plus and minus
    _operator = pyparsing.Char("+") ^ pyparsing.Char("-")
    _operator_factors = immutables.Map({"+": 1, "-": -1})
    _factor_operators = immutables.Map({1: "+", -1: "-"})
    # alphanumeric category codes can be given directly, others have to be quoted
    _category_code = pyparsing.Word(pyparsing.alphanums + ".") ^ pyparsing.QuotedString(
        quoteChar='"', escChar="\\"
    )
    _formula = (
        pyparsing.StringStart()
        + pyparsing.Optional(_operator("unary_op"))
        + _category_code("category_code")
        + pyparsing.ZeroOrMore(_operator("binary_op") + _category_code("category_code"))
        + pyparsing.StringEnd()
    )
    _auxiliary_codes = (
        pyparsing.StringStart()
        + pyparsing.ZeroOrMore(_category_code("aux_category_code"))
        + pyparsing.StringEnd()
    )

    @classmethod
    def _parse_aux_codes(cls, aux_codes_str: str) -> list[str]:
        """Parse a whitespace-separated list of auxiliary codes.

        Parameters
        ----------
        aux_codes_str: str
            Category codes separated by whitespace. Alphanumeric category codes can be
            given directly, other category codes must be quoted using double quotes.

        Returns
        -------
        aux_codes: list
            List of the category codes.

        Examples
        --------
        >>> ConversionRuleSpec._parse_aux_codes("A B")
        ['A', 'B']
        >>> ConversionRuleSpec._parse_aux_codes('"a b" c')
        ['a b', 'c']
        >>> ConversionRuleSpec._parse_aux_codes("")
        []
        >>> ConversionRuleSpec._parse_aux_codes("A + B")
        Traceback (most recent call last):
        ...
        ValueError: Could not parse: 'A + B', error: Expected ...
        """
        try:
            tokens = cls._auxiliary_codes.parseString(aux_codes_str)
        except pyparsing.ParseException as exc:
            raise ValueError(
                f"Could not parse: {aux_codes_str!r}, error: {exc.msg},"
                f" error at char {exc.loc}"
            ) from None
        return list(tokens)

    @classmethod
    def _parse_formula(cls, formula: str) -> dict[str, int]:
        """Parse a formula into factors for categories.

        Parameters
        ----------
        formula: str
            Formula comprising category codes connected with + or - . Alphanumeric
            category codes can be given directly, other category codes must be quoted
            using double quotes.

        Returns
        -------
        code_factors: dict
            mapping of category codes to factors

        Examples
        --------
        >>> ConversionRuleSpec._parse_formula("A + B")
        {'A': 1, 'B': 1}
        >>> ConversionRuleSpec._parse_formula("-A+B")
        {'A': -1, 'B': 1}
        >>> ConversionRuleSpec._parse_formula('"-asdf.#" + B')
        {'-asdf.#': 1, 'B': 1}
        >>> ConversionRuleSpec._parse_formula(" A  -  B")
        {'A': 1, 'B': -1}
        >>> ConversionRuleSpec._parse_formula("-A")
        {'A': -1}
        >>> ConversionRuleSpec._parse_formula('-A+B - "A"')
        {'A': -2, 'B': 1}
        >>> ConversionRuleSpec._parse_formula("-A-")
        Traceback (most recent call last):
        ...
        ValueError: Could not parse: '-A-', error: Expected ...
        >>> ConversionRuleSpec._parse_formula("")
        Traceback (most recent call last):
        ...
        ValueError: Could not parse: '', error: Expected ...
        """
        try:
            tokens = cls._formula.parseString(formula)
        except pyparsing.ParseException as exc:
            raise ValueError(
                f"Could not parse: {formula!r}, error: {exc.msg},"
                f" error at char {exc.loc}"
            ) from None
        # first operator is implicitly a plus, have to handle it specially
        if "unary_op" in tokens:
            op = tokens.pop(0)
        else:
            op = "+"
        code = tokens.pop(0)
        code_factors = {code: cls._operator_factors[op]}
        while len(tokens):
            op = tokens.pop(0)
            code = tokens.pop(0)
            if code in code_factors:
                code_factors[code] += cls._operator_factors[op]
            else:
                code_factors[code] = cls._operator_factors[op]

        return code_factors

    @classmethod
    def from_csv_row(
        cls,
        irow: typing.Iterator[str],
        aux_names: list[str],
        line_number: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
    ) -> "ConversionRuleSpec":
        """Parse a ConversionRuleSpec from a row in a CSV file.

        Parameters
        ----------
        irow: iterable of str
            An iterable (e.g. list) of strings. The first string is the formula for
            the left side (categorization_a), then come the specifications for
            auxiliary categories, with as many fields as there are aux_names, then
            comes the formula for the right side (categorization_b), and finally an
            optional comment.
        aux_names: list of str
            List of names of the auxiliary categorizations.
        line_number: int, optional
            The line number within the CSV, used for nicer error messages if available.

        Returns
        -------
        self: ConversionRuleSpec
            The parsed ConverionRuleSpec.
        """

        n_aux = len(aux_names)
        row = list(irow)

        auxiliary_categories = {}
        factors_a = cls._parse_formula(row[0])
        for i in range(n_aux):
            aux_codes = cls._parse_aux_codes(row[i + 1])
            auxiliary_categories[aux_names[i]] = set(aux_codes)
        factors_b = cls._parse_formula(row[n_aux + 1])

        try:
            comment = row[n_aux + 2]
        except IndexError:
            comment = ""

        return cls(
            factors_categories_a=factors_a,
            factors_categories_b=factors_b,
            auxiliary_categories=auxiliary_categories,
            comment=comment,
            csv_line_number=line_number,
            csv_original_text=",".join(row),
        )

    @classmethod
    def _factors_categories_formula(cls, factors_categories: dict[str, int]) -> str:
        """Serialize a dict mapping categories to factors to a formula string.

        Parameters
        ----------
        factors_categories: dict
            Mapping of categories to factors.

        Returns
        -------
        formula: str
            String representation of the input.
        """
        formula = ""
        first = True
        for category, factor in factors_categories.items():
            while factor != 0:
                if factor > 0:
                    op = "+"
                    factor -= 1
                else:
                    op = "-"
                    factor += 1

                if first:
                    if op == "+":
                        formula += cls._escape_code(category)
                    first = False
                else:
                    formula += f" {op} {cls._escape_code(category)}"
        return formula

    @staticmethod
    def _escape_code(code: str) -> str:
        """Escape a category code for serialization.

        Examples
        --------
        >>> ConversionRuleSpec._escape_code("A")
        'A'
        >>> ConversionRuleSpec._escape_code("2.A")
        '2.A'
        >>> ConversionRuleSpec._escape_code("$1")
        '"$1"'
        >>> ConversionRuleSpec._escape_code('"')
        '"\\\\""'
        """
        if code.isalnum() or code.replace(".", "").isalnum():
            return code
        # replace:
        # \ -> \\
        # " -> \"
        esc = code.replace("\\", "\\\\").replace('"', '\\"')
        return f'"{esc}"'

    def to_csv_row(self) -> list[str]:
        """Return a representation of this rule suitable for writing to a CSV file."""
        row = [self._factors_categories_formula(self.factors_categories_a)]
        for aux_categories in self.auxiliary_categories.values():
            row.append(" ".join(sorted(map(self._escape_code, aux_categories))))
        row.extend(
            (
                self._factors_categories_formula(self.factors_categories_b),
                self.comment,
            )
        )
        return row

    def __str__(self) -> str:
        if self.csv_original_text is not None:
            return self.csv_original_text
        return ",".join(self.to_csv_row())


@dataclasses.dataclass(frozen=True)
class ConversionRule:
    """A rule to convert between categories from two different categorizations.

    Supports one-to-one relationships, one-to-many relationships in both directions and
    many-to-many relationships. For each category, a factor is given which can also be
    negative to model relationships like A = B - C.

    Using auxiliary_categories, a rule can be restricted to specific auxiliary
    categories only.

    Attributes
    ----------
    factors_categories_a : dict mapping categories to factors
        Map of categories from the first categorization to factors. For a simple
        addition, use factor 1, to subtract the category, use factor -1.
    factors_categories_b : dict mapping categories to factors
        Map of categories from the second categorization to factors. For a simple
        addition, use factor 1, to subtract the category, use factor -1.
    auxiliary_categories : dict[Categorization, set[Category]]
        Map of auxiliary categorizations to sets of auxiliary categories. Not
        all auxiliary categorizations need to be specified, and if an auxiliary
        categorization is not specified (or an empty set of category codes is given),
        the validity of the rule is not restricted.
        If an auxiliary categorization is specified and categories are given, the
        rule is only valid for the given categories. If multiple auxiliary
        categorizations are given, the rule is only valid if all auxiliary
        categorizations match.
    comment : str
        A human-readable comment explaining the rule or adding additional information.
    cardinality_a : str
        The cardinality of the rule on side a. Is "one" if there is exactly one category
        in factors_categories_a, and "many" otherwise.
    cardinality_b : str
        The cardinality of the rule on side b. Is "one" if there is exactly one category
        in factors_categories_b, and "many" otherwise.
    is_restricted : bool
        The rule is restricted if and only if for at least one auxiliary categorization
        at least one category is specified, so that the rule is only valid for a
        subset of cases. Otherwise, the rule is unrestricted and valid for all
        cases.
    """

    factors_categories_a: dict["Category", int]
    factors_categories_b: dict["Category", int]
    auxiliary_categories: dict["Categorization", set["Category"]]
    comment: str = ""
    csv_line_number: typing.Optional[int] = None
    csv_original_text: typing.Optional[str] = None
    cardinality_a: str = dataclasses.field(init=False)
    cardinality_b: str = dataclasses.field(init=False)
    is_restricted: bool = dataclasses.field(init=False)

    def __post_init__(self):
        # Have to use object.__setattr__ because the class is frozen. This is fine
        # because we are in __post_init__, so we operate on a not-yet-finished object
        object.__setattr__(
            self,
            "cardinality_a",
            "one" if len(self.factors_categories_a) == 1 else "many",
        )
        object.__setattr__(
            self,
            "cardinality_b",
            "one" if len(self.factors_categories_b) == 1 else "many",
        )
        object.__setattr__(
            self, "is_restricted", any(self.auxiliary_categories.values())
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ConversionRule):
            return NotImplemented
        return (
            self.factors_categories_a == other.factors_categories_a
            and self.factors_categories_b == other.factors_categories_b
            and self.auxiliary_categories == other.auxiliary_categories
            and self.comment == other.comment
        )

    def reversed(self) -> "ConversionRule":
        """Return the ConversionRule with categorization_a and categorization_b
        swapped."""
        return ConversionRule(
            factors_categories_a=self.factors_categories_b,
            factors_categories_b=self.factors_categories_a,
            auxiliary_categories=self.auxiliary_categories,
            comment=self.comment,
            csv_line_number=self.csv_line_number,
            csv_original_text=self.csv_original_text,
        )

    @staticmethod
    def _format_factor_category_human_readable(
        factor: int, category: "Category"
    ) -> str:
        """Format a single category and its factor for humans."""
        if factor == 1:
            return f"{category.categorization.name} {category}"
        else:
            return f"{factor} * {category.categorization.name} {category}"

    def format_human_readable(self, categorization_separator: str = "⮁\n") -> str:
        """Format the rule for humans.

        Parameters
        ----------
        categorization_separator: str, optional
            The categorization_separator is printed between the categories from
            the source categorization and the categories from the target categorization
            to make the difference clear.

        Returns
        -------
        human_readable: str
            The rule in a format optimized for error-free parsing by humans.
        """
        if any(self.auxiliary_categories.values()):
            aux_info = [
                f"{aux_categorization} in {[c.codes[0] for c in sorted(categories)]}"
                for aux_categorization, categories in self.auxiliary_categories.items()
            ]
            r = "Only for " + " and ".join(aux_info) + "\n"
        else:
            r = ""

        r += "\n".join(
            self._format_factor_category_human_readable(f, cat)
            for cat, f in self.factors_categories_a.items()
        )
        r += "\n"
        r += categorization_separator
        r += "\n".join(
            self._format_factor_category_human_readable(f, cat)
            for cat, f in self.factors_categories_b.items()
        )
        r += "\n"

        if self.comment:
            r += f"# Comment: {self.comment!r}\n"

        return r

    def to_spec(self) -> ConversionRuleSpec:
        """Return a serializable specification.

        Returns
        -------
        spec: ConversionRuleSpec
        """
        return ConversionRuleSpec(
            factors_categories_a={
                category.codes[0]: factor
                for category, factor in self.factors_categories_a.items()
            },
            factors_categories_b={
                category.codes[0]: factor
                for category, factor in self.factors_categories_b.items()
            },
            auxiliary_categories={
                categorization.name: {category.codes[0] for category in categories}
                for categorization, categories in self.auxiliary_categories.items()
            },
            comment=self.comment,
            csv_line_number=self.csv_line_number,
            csv_original_text=self.csv_original_text,
        )

    def __str__(self):
        return str(self.to_spec())

    def format_with_lineno(self) -> str:
        """Human-readable string representation of the rule with information in which
        line in the CSV file it was defined, if that is available."""
        s = f"<Rule '{self!s}'"
        if self.csv_line_number is not None:
            s += f" from line {self.csv_line_number}"
        s += ">"
        return s


class ConversionBase:
    """Common base of ConversionSpec and Conversion.

    Mainly used to hold a single definition of the metadata attributes

    Attributes
    ----------
    categorization_a_name : str
        Name of the first categorization.
    categorization_b_name : str
        Name of the second categorization.
    auxiliary_categorizations_names : list of str, optional
        Names of the auxiliary categorizations.
    rule_specs : list of ConversionRuleSpec
        The rule specifications for conversion between individual categories or sets of
        categories.
    comment : str, optional
        Notes and explanations for humans.
    references : str, optional
        Citable reference(s) for the conversion.
    institution : str, optional
        Where the conversion originates.
    last_update : datetime.date, optional
        The date of the last change.
    version : str, optional
        The version of the ConversionRules, if there are multiple versions.
    """

    def __init__(
        self,
        *,
        categorization_a_name: str,
        categorization_b_name: str,
        rule_specs: list[ConversionRuleSpec],
        auxiliary_categorizations_names: typing.Optional[list[str]] = None,
        comment: typing.Optional[str] = None,
        references: typing.Optional[str] = None,
        institution: typing.Optional[str] = None,
        last_update: typing.Optional[datetime.date] = None,
        version: typing.Optional[str] = None,
    ):
        self.categorization_a_name = categorization_a_name
        self.categorization_b_name = categorization_b_name
        self.rule_specs = rule_specs
        self.auxiliary_categorizations_names = auxiliary_categorizations_names
        self.comment = comment
        self.references = references
        self.institution = institution
        self.last_update = last_update
        self.version = version


class ConversionSpec(ConversionBase):
    """Specification of rules for conversion between two categorizations, with support
    for alternative rules depending on auxiliary categorizations.

    This class supports parsing the rules from a specification file and other
    operations which can be performed on the pure rules without knowledge of the
    categorization objects themselves.

    Attributes
    ----------
    categorization_a_name : str
        Name of the first categorization.
    categorization_b_name : str
        Name of the second categorization.
    auxiliary_categorizations_names : list of str, optional
        Names of the auxiliary categorizations.
    rule_specs : list of ConversionRuleSpec
        The rule specifications for conversion between individual categories or sets of
        categories.
    comment : str, optional
        Notes and explanations for humans.
    references : str, optional
        Citable reference(s) for the conversion.
    institution : str, optional
        Where the conversion originates.
    last_update : datetime.date, optional
        The date of the last change.
    version : str, optional
        The version of the ConversionRules, if there are multiple versions.
    """

    _meta_data_keys: typing.ClassVar[list[str]] = [
        "comment",
        "references",
        "institution",
        "last_update",
        "version",
    ]

    _strictyaml_metadata_schema = sy.Map(
        {
            sy.Optional("comment"): sy.Str(),
            sy.Optional("references"): sy.Str(),
            sy.Optional("institution"): sy.Str(),
            sy.Optional("last_update"): sy.Datetime(),
            sy.Optional("version"): sy.Str(),
        }
    )

    def __init__(
        self,
        *,
        categorization_a_name: str,
        categorization_b_name: str,
        rule_specs: list[ConversionRuleSpec],
        auxiliary_categorizations_names: typing.Optional[list[str]] = None,
        comment: typing.Optional[str] = None,
        references: typing.Optional[str] = None,
        institution: typing.Optional[str] = None,
        last_update: typing.Optional[datetime.date] = None,
        version: typing.Optional[str] = None,
    ):
        ConversionBase.__init__(
            self,
            categorization_a_name=categorization_a_name,
            categorization_b_name=categorization_b_name,
            rule_specs=rule_specs,
            auxiliary_categorizations_names=auxiliary_categorizations_names,
            comment=comment,
            references=references,
            institution=institution,
            last_update=last_update,
            version=version,
        )

    @classmethod
    def _read_csv_meta(cls, fd: typing.TextIO):
        """Read the metadata section of a CSV conversion specification file. It consists
        of YAML key, value pairs, one pair on each line separated by a colon.
        Each line is prefixed with the comment char "#".

        Parameters
        ----------
        fd: a CSV file object
            Use a file object which was not used before to read from.
            The file object will be iterated up to the end of the meta data
            section, so that after calling _read_csv_meta you can directly
            start reading the data section.

        Returns
        -------
        meta_data: dict
            Mapping of meta data keys to values.
        linecount: int
            Count of lines of the metadata block
        """
        yaml_header = ""
        last_pos = fd.tell()
        line = fd.readline()
        while line.startswith("#"):
            # remove leading comment and whitespace
            yaml_header += line[1:].lstrip()
            last_pos = fd.tell()
            line = fd.readline()
        fd.seek(last_pos)
        meta_data = sy.load(yaml_header, schema=cls._strictyaml_metadata_schema).data

        return meta_data, yaml_header.count("\n")

    @classmethod
    def _read_csv_rules(
        cls, reader: csv.reader, offset: int
    ) -> tuple[str, str, list[str], list[ConversionRuleSpec]]:
        """Read the data section of a CSV specification file. It consists of a header,
        followed by rules, with each rule on one line.

        Parameters
        ----------
        reader: CSV reader object as returned by csv.reader
            The reader object must already be advanced to the rules section, so that
            the first read yields the data header.
        offset: int
            Number of lines of the metadata block.

        Returns
        -------
        a_name, b_name, aux_names, rule_specs: str, str, list, list
           The name of categorizations A and B, the names of the auxiliary categories,
           and the parsed rules.
        """
        rule_specs = []
        header: list[str] = next(reader)
        a_name = header[0]
        b_name = header[-2]
        if header[-1] != "comment":
            raise ValueError("Last column must be 'comment', but isn't.")
        aux_names = header[1:-2]
        for row in reader:
            line_num = reader.line_num + offset
            irow = iter(row)
            try:
                rule_specs.append(
                    ConversionRuleSpec.from_csv_row(
                        irow, aux_names=aux_names, line_number=line_num
                    )
                )
            except ValueError as err:
                raise ValueError(f"Error in line {line_num}: {err}") from None

        return a_name, b_name, aux_names, rule_specs

    @classmethod
    def _from_csv(
        cls,
        fd: typing.TextIO,
    ) -> "ConversionSpec":
        meta_data, len_meta_data = cls._read_csv_meta(fd)
        reader = csv.reader(fd, quoting=csv.QUOTE_NONE, escapechar="\\")
        a_name, b_name, aux_names, rule_specs = cls._read_csv_rules(
            reader, len_meta_data
        )

        return cls(
            categorization_a_name=a_name,
            categorization_b_name=b_name,
            rule_specs=rule_specs,
            auxiliary_categorizations_names=aux_names or None,
            **meta_data,
        )

    @classmethod
    def from_csv(
        cls,
        filepath: typing.Union[str, pathlib.Path, typing.TextIO],
    ) -> "ConversionSpec":
        """Read conversion from comma-separated-values file."""
        if not isinstance(filepath, (str, pathlib.Path)):
            return cls._from_csv(filepath)
        fp = pathlib.Path(filepath)
        with fp.open(newline="") as fd:
            return cls._from_csv(fd)

    def __repr__(self):
        return (
            f"<ConversionSpec {self.categorization_a_name!r} <->"
            f" {self.categorization_b_name!r} with {len(self.rule_specs)} rules>"
        )

    def hydrate(
        self,
        cats: dict[str, "Categorization"],
    ) -> "Conversion":
        """Convert this Specification into a Conversion object with full
        functionality."""
        categorization_a = cats[self.categorization_a_name]
        categorization_b = cats[self.categorization_b_name]
        auxiliary_categorizations = (
            [cats[x] for x in self.auxiliary_categorizations_names]
            if self.auxiliary_categorizations_names
            else None
        )
        return Conversion(
            categorization_a=categorization_a,
            categorization_b=categorization_b,
            rules=[
                rule_spec.hydrate(
                    categorization_a=categorization_a,
                    categorization_b=categorization_b,
                    cats=cats,
                )
                for rule_spec in self.rule_specs
            ],
            auxiliary_categorizations=auxiliary_categorizations,
            comment=self.comment,
            references=self.references,
            institution=self.institution,
            last_update=self.last_update,
            version=self.version,
        )


@dataclasses.dataclass(frozen=True)
class OverCountingProblem:
    """A suspected over counting problem."""

    category: "HierarchicalCategory"
    leave_node_groups: list[set["HierarchicalCategory"]]
    rules: list[ConversionRule]

    def __str__(self):
        involved_rules_str = ", ".join(rule.format_with_lineno() for rule in self.rules)
        sorted_leave_node_groups = [sorted(g) for g in self.leave_node_groups]
        return (
            f"{self.category!r} is possibly counted multiple times"
            f"\ninvolved leave groups categories: {sorted_leave_node_groups!r}"
            f"\ninvolved rules: {involved_rules_str}."
        )


class Conversion(ConversionBase):
    """Conversion between two categorizations.

    This class collects functionality which needs access to the actual categorizations
    and categories.

    Attributes
    ----------
    categorization_a : Categorization
        The first categorization.
    categorization_b : Categorization
        The second categorization.
    auxiliary_categorizations : list of Categorization, optional
        The auxiliary categorizations, if any.
    rules : list of ConversionRule
        The actual rules for conversion between individual categories or sets of
        categories.
    comment : str, optional
        Notes and explanations for humans.
    references : str, optional
        Citable reference(s) for the conversion.
    institution : str, optional
        Where the conversion originates.
    last_update : datetime.date, optional
        The date of the last change.
    version : str, optional
        The version of the ConversionRules, if there are multiple versions.
    """

    def __init__(
        self,
        *,
        categorization_a: "Categorization",
        categorization_b: "Categorization",
        rules: list[ConversionRule],
        auxiliary_categorizations: typing.Optional[list["Categorization"]] = None,
        comment: typing.Optional[str] = None,
        references: typing.Optional[str] = None,
        institution: typing.Optional[str] = None,
        last_update: typing.Optional[datetime.date] = None,
        version: typing.Optional[str] = None,
    ):
        ConversionBase.__init__(
            self,
            categorization_a_name=categorization_a.name,
            categorization_b_name=categorization_b.name,
            rule_specs=[rule.to_spec() for rule in rules],
            auxiliary_categorizations_names=[x.name for x in auxiliary_categorizations]
            if auxiliary_categorizations
            else None,
            comment=comment,
            references=references,
            institution=institution,
            last_update=last_update,
            version=version,
        )
        self.categorization_a = categorization_a
        self.categorization_b = categorization_b
        self.rules = rules
        self.auxiliary_categorizations = auxiliary_categorizations

    @staticmethod
    def from_csv(
        filepath: typing.Union[str, pathlib.Path, typing.TextIO],
        cats: typing.Union[dict[str, "Categorization"], None] = None,
    ) -> "Conversion":
        """Read conversion from comma-separated-values file and add categorizsations."""

        conv = ConversionSpec.from_csv(filepath)

        if cats is None:
            import climate_categories

            cats = climate_categories.cats

        return conv.hydrate(cats=cats)

    def reversed(self) -> "Conversion":
        """Returns the Conversion with categorization_a and categorization_b swapped."""
        return Conversion(
            categorization_a=self.categorization_b,
            categorization_b=self.categorization_a,
            rules=[rule.reversed() for rule in self.rules],
            auxiliary_categorizations=self.auxiliary_categorizations,
            comment=self.comment,
            references=self.references,
            institution=self.institution,
            last_update=self.last_update,
            version=self.version,
        )

    def __repr__(self):
        return (
            f"<Conversion {self.categorization_a_name!r} <->"
            f" {self.categorization_b_name!r} with {len(self.rule_specs)} rules>"
        )

    def describe_detailed(self) -> str:
        """Detailed human-readable description of the conversion rules.

        Sections are added for direct one-to-one mappings, one-to-many mappings,
        many-to-one mappings, and many-to-many mappings, respectively.

        Factors are shown at the start of the line if they don't equal 1, like this:
        -1 * IPCC1996 4 Agriculture
        to indicate that category 4 should be subtracted.
        """
        one_to_one = []
        one_to_many = []
        many_to_one = []
        many_to_many = []
        cats_a: set[Category] = set()
        cats_b: set[Category] = set()
        for rule in self.rules:
            cats_a.update(rule.factors_categories_a.keys())
            cats_b.update(rule.factors_categories_b.keys())
            if rule.cardinality_a == "one" and rule.cardinality_b == "one":
                one_to_one.append(rule)
            elif rule.cardinality_a == "one":
                one_to_many.append(rule)
            elif rule.cardinality_b == "one":
                many_to_one.append(rule)
            else:
                many_to_many.append(rule)

        cat_a, cat_b = self.categorization_a.name, self.categorization_b.name

        r = f"# Mapping between {cat_a} and {cat_b}\n\n"
        r += "## Simple direct mappings\n\n"
        r += "\n".join(
            rule.format_human_readable(categorization_separator="")
            for rule in one_to_one
        )
        r += "\n\n"
        r += f"## One-to-many mappings - one {cat_a} to many {cat_b}\n\n"
        r += "\n".join((rule.format_human_readable()) for rule in one_to_many)
        r += "\n\n"
        r += f"## Many-to-one mappings - many {cat_a} to one {cat_b}\n\n"
        r += "\n".join((rule.format_human_readable()) for rule in many_to_one)
        r += "\n\n"
        r += f"## Many-to-many mappings - many {cat_a} to many {cat_b}\n\n"
        r += "\n".join((rule.format_human_readable()) for rule in many_to_many)
        r += "\n\n"

        r += "## Unmapped categories\n\n"
        cats_missing_a = set(self.categorization_a.values()) - cats_a
        cats_missing_b = set(self.categorization_b.values()) - cats_b
        r += f"### {cat_a}\n"
        r += "\n".join(sorted(str(x) for x in cats_missing_a)) + "\n\n"
        r += f"### {cat_b}\n"
        r += "\n".join(sorted(str(x) for x in cats_missing_b)) + "\n\n"

        return r

    def find_unmapped_categories(
        self,
    ) -> tuple[set["Category"], set["Category"]]:
        """Find categories for which no rule exists to map them.

        Returns
        -------
        missing_categories_a, missing_categories_b: set, set
            A list of categories missing from categorization_a and categorization_b,
            respectively.
        """
        cats_a: set[Category] = set()
        cats_b: set[Category] = set()
        for rule in self.rules:
            cats_a.update(rule.factors_categories_a.keys())
            cats_b.update(rule.factors_categories_b.keys())

        cats_missing_a = set(self.categorization_a.values()) - cats_a
        cats_missing_b = set(self.categorization_b.values()) - cats_b
        return cats_missing_a, cats_missing_b

    def find_over_counting_problems(self) -> list[OverCountingProblem]:
        """Check if any category from one side is counted more than once on the
        other side.

        Note that the algorithm at the moment can't reliably detect all over counting
        problems and also some suspected problems might be fine under closer
        examination, so use this function only to generate hints for possible problems.

        Returns
        -------
        problems: list of OverCountingProblem objects
            All detected suspected problems.
        """
        for categorization in self.categorization_a, self.categorization_b:
            if not categorization.hierarchical:
                raise ValueError(
                    f"{categorization} is not hierarchical, without "
                    f"a hierarchy, over counting can not be evaluated."
                )
            if not categorization.total_sum:  # type: ignore
                raise ValueError(
                    f"For {categorization} it is not specified that the "
                    f"sum of a set of children equals the parent, so "
                    f"over counting can not be evaluated."
                )

        problems = []
        for categorization in self.categorization_a, self.categorization_b:
            # used to cache costly descendant evaluation
            descendants: dict[str, set[str]] = {}
            for category in categorization.values():
                prob = self._check_over_counting_category(
                    category,
                    categorization,
                    descendants,  # type: ignore
                )
                if prob:
                    problems.append(prob)

        return problems

    @staticmethod
    def _leave_node_group(
        categories: typing.Iterable["HierarchicalCategory"],
        hull: set[str],
        descendants: dict[str, set[str]],
    ) -> bool:
        """Are all of the given categories leave nodes of the given hull?

        Parameters
        ----------
        categories: list of HierarchicalCategory objects
            Categories that will be checked. If any of the categories has descendants
            outside of the hull, the function will return False.
        hull: set of strings
            Set of primary codes of HierarchicalCategories, which define the hull
            that will be used to check the categories.
        descendants: dict[str, list[str]]
            Mapping of primary codes of parent HierarchicalCategories to the codes
            of their descendants. Will be filled with additional mappings if they are
            computed. Re-use the dictionary for better performance.

        Returns
        -------
        all_leave: bool
            If all categories are leave categories within the given hull, returns True.
            Otherwise, returns false.
        """
        for c in categories:
            # Use cached descendants information if it is available, compute and cache
            # it otherwise
            try:
                desc = descendants[c.codes[0]]
            except KeyError:
                desc = {d.codes[0] for d in c.descendants}
                descendants[c.codes[0]] = desc

            for d in desc:
                if d in hull:
                    return False
        return True

    def relevant_rules(
        self,
        categories: set["HierarchicalCategory"],
        source_categorization: typing.Optional["Categorization"] = None,
        simple_sums_only: bool = False,
    ) -> list[ConversionRule]:
        """Returns all rules which involve the given categories.

        Parameters
        ----------
        categories: set of HierarchicalCategory
            The categories to limit the rules to.
        source_categorization: Categorization, optional
            The categorization that the categories are part of, either
            self.categorization_a or self.categorization_b.
        simple_sums_only: bool, default False
            If true, only consider rules where the given categories enter as simple
            summands (i.e. with a factor of 1).

        Returns
        -------
        relevant_rules:
            All rules which touch the given categories.
        """
        relevant_rules: list[ConversionRule] = []
        if not categories:
            return relevant_rules

        if source_categorization is None:
            source_categorization = next(iter(categories)).categorization

        for rule in self.rules:
            if source_categorization == self.categorization_a:
                fc = rule.factors_categories_a
            else:
                fc = rule.factors_categories_b

            if simple_sums_only:
                rule_source_categories = {
                    cat for cat, factor in fc.items() if factor == 1
                }
            else:
                rule_source_categories = {cat for cat, factor in fc.items()}

            if categories.intersection(rule_source_categories):
                relevant_rules.append(rule)

        return relevant_rules

    def _check_over_counting_category(
        self,
        category: "HierarchicalCategory",
        source_categorization: "Categorization",
        descendants: dict[str, set[str]],
    ) -> typing.Optional[OverCountingProblem]:
        """Finds possible over counting problems for the specified category.

        Parameters
        ----------
        category: HierarchicalCategory
            The category to check.
        source_categorization: Categorization
            The categorization which contains the category (either self.categorization_a
            or self.categorization_b).
        descendants: dict
            Caching dict with descendant information. Before calculating potentially
            costly descendant information, it will be taken from this dict. If new
            descendant information is calculated, it will be put into this dict.

        Notes
        -----
        The algorithm is:

        Definition:
        The ancestral set A(c) of a category c is the set
        comprising the category, its parents, and all members of the ancestral set of
        each of its parents.

        Definition:
        The descendents D(c) of a category c are the children of c, and all
        descendents of the children of c.

        Definition:
        The projection P_S(c) of a category c using the conversion S is the set of
        categories which receive at least a part of the contents of category c according
        to the rules of conversion S.
        I assume without loss of generality that c is part of the left-hand-side
        Categorization of S.
        Then, assuming that the conversion S only contains simple sums
        without repeated categories in its rules, it follows that P_S(c) is the set of
        categories which are in the right hand side of rules in S where c is in the
        left hand side.

        Definition:
        The ancestral projections PA_S(c) of a category c using the conversion S is the
        set of projections of the ancestral set, i.e.
        PA_S(c) = {P_S(a) for a in A(c)}

        Definition:
        The hull hull(MM) of the set MM, which is comprised of sets itself, is the union
        of all members of MM.

        Definition:
        A largest element max(MM) of the set MM, which is comprised of sets itself, is
        an element with the highest number of elements.

        Definition:
        The leave node groups L(MM) of a set MM, which is comprised of sets itself, are
        the sets in MM which have only members that have no descendant in hull(MM).
        L(MM) = {M in MM and
           (for all members c of M:
               for all descendants d of c:
                d not in hull(MM)
           )}

        Then, an over counting problem is found for category c if
        hull(L(PA_S(c))) != max(L(PA_S(C)))
        """

        # A(c)
        ancestral_set = set(category.ancestors)
        ancestral_set.add(category)

        # PA_S(c)
        relevant_rules = self.relevant_rules(
            categories=ancestral_set,
            source_categorization=source_categorization,
            simple_sums_only=True,
        )
        # TODO: for now, only use rules that don't have aux categories
        relevant_rules = [rule for rule in relevant_rules if not rule.is_restricted]
        projected_ancestral_set: list[set[HierarchicalCategory]] = []
        for rule in relevant_rules:
            if source_categorization == self.categorization_a:
                fc = rule.factors_categories_b
            else:
                fc = rule.factors_categories_a
            target_categories = {cat for cat, factor in fc.items() if factor == 1}
            projected_ancestral_set.append(target_categories)  # type: ignore

        if not projected_ancestral_set:  # trivial
            return None

        # for performance, use codes (which are guaranteed to be unique within a
        # categorization) for the comparisons here
        projected_ancestral_set_codes = [
            {c.codes[0] for c in group} for group in projected_ancestral_set
        ]

        # hull(PA_S(c))
        hull: set[str] = set().union(*projected_ancestral_set_codes)

        # L(PA_S(c))
        leave_node_groups = [
            m
            for m in projected_ancestral_set
            if self._leave_node_group(m, hull, descendants)
        ]

        leave_hull = set().union(*leave_node_groups)
        largest = max(leave_node_groups, key=len)

        if len(leave_hull) != len(largest):
            return OverCountingProblem(
                category=category,
                rules=relevant_rules,
                leave_node_groups=leave_node_groups,
            )
        else:
            return None

    def __eq__(self, other):
        return (
            isinstance(other, Conversion)
            and self.categorization_a == other.categorization_a
            and self.categorization_b == other.categorization_b
            and self.rules == other.rules
        )



import climate_categories


def test_number_of_categories():
    assert len(climate_categories.RCMIP) == 87  # 86 from Excel sheet + top level


def test_levels():
    assert climate_categories.RCMIP.level("Emissions") == 1
    assert climate_categories.RCMIP.level("Emissions|F-Gases|HFC|HFC236fa") == 4



"""Run this via `make climate_categories/data/IPCC1996.yaml` in the main directory."""

import pathlib
import typing

import camelot
from utils import download_cached, title_case

import climate_categories

URL = "https://www.ipcc-nggip.iges.or.jp/public/gl/guidelin/ch1ri.pdf"
INPATH = pathlib.Path("./data_generation/IPCC1996.pdf")
OUTPATH = pathlib.Path("./climate_categories/data/IPCC1996.yaml")


def parse_pages(fpath, pages, areas, columns) -> list:
    ts = []
    for page in pages:
        t = camelot.read_pdf(
            str(fpath),
            pages=str(page),
            flavor="stream",
            table_areas=[areas[page % 2]],
            columns=[columns[page % 2]],
            row_tol=10,
        )
        ts.append(t[0])
    return ts


def read_pdf(fpath):
    # 1 energy
    ts = []
    ts += parse_pages(
        fpath,
        pages=range(3, 8),
        areas=["168,723,552,120", "43,723,468,114"],
        columns=["279", "194"],
    )
    # 2 industrial processes
    # 3 solvent and other product use
    ts += parse_pages(
        fpath,
        pages=range(8, 11),
        areas=["140,723,541,135", "0,723,398,117"],
        columns=["285", "206"],
    )
    # 4 agriculture
    ts += parse_pages(
        fpath,
        pages=range(11, 14),
        areas=["157,723,546,95", "55,723,470,88"],
        columns=["276", "193"],
    )
    # 5 agriculture
    ts += parse_pages(
        fpath,
        pages=range(14, 17),
        areas=["127,723,550,214", "55,723,466,208"],
        columns=["276", "198"],
    )
    # 6 waste
    ts.append(
        camelot.read_pdf(
            str(fpath),
            pages="17",
            flavor="stream",
            table_areas=["52,723,464,159"],
            columns=["210"],
            row_tol=10,
        )[0]
    )
    # 7 other
    ts.append(
        camelot.read_pdf(
            str(fpath),
            pages="18",
            flavor="stream",
            table_areas=["180,723,552,670"],
            columns=["288"],
            row_tol=10,
        )[0]
    )
    return ts


def combine_rows(ts):
    rows = []
    for table in ts:
        code_name_previously_empty = True
        definition_previously_empty = True
        current_code_name = None
        current_definition = None
        for _, (code_name, definition) in table.df.iterrows():
            if (
                code_name.startswith("Time period is an important element in")
                or code_name.startswith("categories.  For example, the IPCC default")
                or code_name.startswith("years for biomass decay.")
            ):
                continue
            code_name_empty = not code_name.strip()
            definition_empty = not definition.strip()

            if (
                (code_name_previously_empty and not code_name_empty)
                or (definition_previously_empty and not definition_empty)
                or "\n" in code_name
                or code_name
                in (
                    "iv Motorcycles",
                    "Mining activities",
                    "2 D OTHER PRODUCTION",
                    "2 G OTHER",
                    "4 B MANURE",
                    "6 B WASTEWATER",
                    "6 C WASTE INCINERATION",
                )
            ) and not (
                current_code_name in (" 2\nINDUSTRIAL",)
                or definition.startswith(
                    "petroleum  products,  the",
                )
            ):
                # a new logical row started
                if current_code_name is not None:
                    rows.append((current_code_name, current_definition))
                current_code_name = ""
                current_definition = ""

            if code_name:
                current_code_name += f" {code_name}"
            if definition:
                current_definition += f" {definition}"

            code_name_previously_empty = code_name_empty
            definition_previously_empty = definition_empty

        rows.append((current_code_name, current_definition))

    return rows


def parse_codes(rows):
    cats_raw = []
    roman_numerals = ("i", "ii", "iii", "iv", "v", "vi", "vii")
    parent_code = ""
    parent_roman_code = ""
    running_number = 1
    for code_title, definition in rows:
        if code_title.startswith(" 1 A 2\nb "):
            code_raw = code_title[:8]
            title = code_title[9:]
        elif code_title == " 1 A 2\ne":
            code_raw = code_title
            title = definition
            definition = ""
        elif code_title == " 1 A 3\nd Navigation":
            code_raw = "1 A 3 d"
            title = "Navigation"
        elif code_title == " 4G\nOTHER":
            code_raw = "4 G"
            title = "OTHER"
        elif (
            code_title.startswith(" 1 A 4\n")
            or code_title.startswith(" 1 A 5\nb")
            or code_title.startswith(" 1 B 1\na")
        ) and code_title[7] in ("a", "b", "c"):
            code_raw = code_title[:8]
            title = code_title[9:]
        elif code_title == " FUGITIVE EMISSIONS FROM FUELS":
            code_raw = "1 B"
            title = code_title
        elif (
            "\n" in code_title
            and code_title.strip()[0].isnumeric()
            and code_title
            not in (
                " 4 B\nANAEROBIC 10",
                " 5 A 1\na Wet/ very moist",
                " 5 A 1\nh Other",
                " 5 A 2\nd Other",
            )
        ):
            code_raw, title = code_title.rsplit("\n", 1)
        else:
            splitted = code_title.split()
            if len(splitted) == 2 and len(splitted[0]) == 1:
                code_raw, title = splitted
            else:
                code_raw = ""
                title = ""
                for s in splitted:
                    if len(s) == 1 or s in roman_numerals or s.isnumeric():
                        code_raw += f" {s}"
                    else:
                        title += f" {s}"

        if code_raw == " 2 B3":
            code = "2.B.3"
        else:
            code = ".".join(code_raw.split())
        if code in roman_numerals:
            code = f"{parent_code}.{code}"
            running_number = 1
            parent_roman_code = code
        elif code in ("c.i", "c.ii"):
            code = f"{parent_code[:-1]}{code}"
            running_number = 1
            parent_roman_code = code
        elif not code:
            code = f"{parent_roman_code}.{running_number}"
            running_number += 1
        else:
            parent_code = code
            running_number = 1

        cats_raw.append((code, title, definition))

    return cats_raw


def parse_categories(cats_raw):
    cats = {}
    for code, title, definition in cats_raw:
        if "(" in title and "ISIC" in title and not definition:
            title, definition = title.split("(")
            definition = definition[:-1]

        title = title_case(" ".join(title.split()))
        definition = " ".join(definition.split()).replace("- ", "")
        if len(definition) > 2 and definition[0] == "(" and definition[-1] == ")":
            definition = definition[1:-1]

        altcode = code.replace(".", "")
        altcode2 = code.replace(".", " ")
        spec = {
            "title": title,
        }
        if definition and definition.strip() != ".":
            spec["comment"] = definition
        if altcode != code:
            spec["alternative_codes"] = [altcode, altcode2]

        if code in cats:
            raise ValueError(f"{code!r} exists twice")
        cats[code] = spec

    return cats


def add_relationships(categories):
    for parent_code in categories:
        prefix = f"{parent_code}."
        children = []
        for child_code in categories:
            if child_code.startswith(prefix) and "." not in child_code[len(prefix) :]:
                children.append(child_code)
        if children:
            categories[parent_code]["children"] = [children]


def main():
    download_cached(URL, INPATH)
    ts = read_pdf(INPATH)
    rows = combine_rows(ts)
    cats_raw = parse_codes(rows)

    # Widely used and very useful, even though not included in the official spec
    categories: dict[str, typing.Any] = {
        "0": {
            "title": "National Total",
            "comment": "All emissions and removals",
            "children": [["1", "2", "3", "4", "5", "6", "7"]],
        }
    }

    categories.update(parse_categories(cats_raw))
    add_relationships(categories)

    spec = {
        "name": "IPCC1996",
        "title": "IPCC GHG emission categories (1996)",
        "comment": "IPCC classification of green-house gas emissions and removals into"
        " categories, 1996 edition",
        "references": "IPCC 1996, Revised 1996 IPCC Guidelines for National Greenhouse"
        " Gas Inventories: Reporting Instructions,"
        " Volume 1, Chapter 1.1, pages 1.2ff,"
        " https://www.ipcc-nggip.iges.or.jp/public/gl/guidelin/ch1ri.pdf",
        "institution": "IPCC",
        "last_update": "1996-09-13",
        "hierarchical": True,
        "version": "1996",
        "total_sum": True,
        "categories": categories,
        "canonical_top_level_category": "0",
    }

    categories = spec["categories"]
    categories["1.A.1"]["comment"] = categories["1.A.1"]["comment"].replace(
        "energyproducing", "energy producing"
    )

    categories["1.A.2"]["title"] = "Manufacturing Industries and Construction"
    categories["1.A.2"]["comment"] = categories["1.A.2"]["comment"].replace(
        "Emissions from the industry sector should be specified by subsectors that "
        "correspond to the International Standard Industrial Classification of All "
        "Economic Activities (ISIC).",
        "Emissions from the industry sector should be specified by subsectors that "
        "correspond to the International Standard Industrial Classification of All "
        "Economic Activities, 3rd Edition (ISIC) "
        " [International Standard Industrial Classification of all Economic Activities,"
        " Series M No. 4, Rev. 3, United Nations, New York, 1990].",
    )

    categories["1.A.5"]["title"] = "Other"

    categories["2.C.4"]["title"] = "SF6 used in Aluminium and Magnesium Foundries"

    categories["4.B.10"]["title"] = "Anaerobic Lagoons"

    categories["4.C.3.a"]["title"] = "Water Depth 50-100 cm"
    categories["4.C.3.b"]["title"] = "Water Depth > 100 cm"

    categories["5.A.5"]["comment"] = (
        "Emissions and removals of CO2 from other biomass categories, including"
        " village and farm trees, etc."
        " This category"
        " is intended to account for biomass which is found in locations other than"
        " the major ecosystem types listed. This includes dispersed trees in villages,"
        " farms, urban areas, etc., and also includes additional ecosystem types which"
        " may be important for biomass accounting in specific countries."
        " Afforestation programmes which create forests will be accounted for in the"
        " appropriate forest ecosystem category. Afforestation which produces"
        " dispersed trees, e.g., urban tree planting, would be accounted for in"
        " “Other.”"
    )

    categories["5.D"]["title"] = "CO2 Emissions and Removals From Soil"

    IPCC1996 = climate_categories.HierarchicalCategorization.from_spec(spec)

    IPCC1996.to_yaml(OUTPATH)

    climate_categories.HierarchicalCategorization.from_yaml(OUTPATH)


if __name__ == "__main__":
    main()



from collections.abc import Iterable

from . import _categories


def search_code(
    code: str, cats: Iterable[_categories.Categorization]
) -> set[_categories.Category]:
    """Search for the given code in the given categorizations."""
    return {cat[code] for cat in cats if code in cat.all_keys()}



"""Run this via `make climate_categories/data/ISO3.yaml` in the main directory."""

import pathlib
import typing

import requests

import climate_categories

URL = (
    "https://salsa.debian.org/iso-codes-team/iso-codes/-/raw/main/data/"
    "iso_3166-1.json?inline=false"
)
OUTPATH = pathlib.Path("./climate_categories/data/ISO3.yaml")


def main():
    """Generate the categorization."""

    categories = load_countries()

    # add some widely used additional categories
    categories["World"] = {
        "title": "The world",
        "alternative_codes": ["EARTH", "Earth", "WORLD"],
        "children": [list(categories.keys())],
    }
    categories = add_eu_categories(categories)
    categories = add_unfccc_categories(categories)
    categories = add_unfccc_names(categories)
    categories = add_aosis(categories)
    categories = add_g7g20(categories)
    categories = add_oecd(categories)
    categories = add_historical_names(categories)

    spec = {
        "name": "ISO3",
        "title": "ISO 3166-1 countries with climate-relevant groupings",
        "comment": "Countries, regions, and other areas. Also includes information on "
        "groups like being included in Annex I of the UN Framework Convention on "
        "Climate Change.",
        "references": """ISO 3166, https://www.iso.org/iso-3166-country-codes.html;
iso-codes package, https://salsa.debian.org/iso-codes-team/iso-codes;
UNFCCC Parties & Observers, https://unfccc.int/parties-observers;
EU members,
https://ec.europa.eu/eurostat/statistics-explained/index.php?title=Glossary:EU_enlargements;
G7 and G20, https://www.bmuv.de/themen/europa-internationales/internationales/g7-und-g20;
OECD members, https://www.oecd.org/about/document/ratification-oecd-convention.htm;
AOSIS members, https://www.aosis.org/about/member-states/""",
        "institution": "UN",
        "last_update": "2023-06-22",
        "hierarchical": True,
        "version": "2023-06-22",
        "total_sum": False,
        "categories": categories,
        "canonical_top_level_category": "WORLD",
    }

    return climate_categories.HierarchicalCategorization.from_spec(spec)


def add_aosis(categories):
    categories["AOSIS"] = {
        "title": "Alliance of Small Island States",
        "children": [
            [
                "ATG",
                "BHS",
                "BRB",
                "BLZ",
                "CUB",
                "DMA",
                "DOM",
                "GRD",
                "GUY",
                "HTI",
                "JAM",
                "KNA",
                "LCA",
                "VCT",
                "SUR",
                "TTO",
                "COK",
                "FSM",
                "FJI",
                "KIR",
                "NRU",
                "NIU",
                "PLW",
                "PNG",
                "MHL",
                "WSM",
                "SLB",
                "TON",
                "TUV",
                "VUT",
                "CPV",
                "COM",
                "GNB",
                "MDV",
                "MUS",
                "STP",
                "SYC",
                "SGP",
                "TLS",
            ]
        ],
    }
    return categories


def add_historical_names(categories):
    categories["TUR"]["info"]["historical_names"] = ["Turkey"]
    return categories


def add_oecd(categories):
    categories["OECD"] = {
        "title": "Organisation for Economic Co-operation and Development",
        "children": [
            [
                "AUS",
                "AUT",
                "BEL",
                "CAN",
                "CHL",
                "COL",
                "CRI",
                "CZE",
                "DNK",
                "EST",
                "FIN",
                "FRA",
                "DEU",
                "GRC",
                "HUN",
                "ISL",
                "IRL",
                "ISR",
                "ITA",
                "JPN",
                "KOR",
                "LVA",
                "LTU",
                "LUX",
                "MEX",
                "NLD",
                "NZL",
                "NOR",
                "POL",
                "PRT",
                "SVK",
                "SVN",
                "ESP",
                "SWE",
                "CHE",
                "TUR",
                "GBR",
                "USA",
            ]
        ],
    }
    return categories


def add_g7g20(categories):
    categories["G7"] = {
        "title": "Group of Seven",
        "children": [["DEU", "FRA", "GBR", "ITA", "JPN", "USA", "CAN", "EU"]],
    }
    categories["G8"] = {
        "title": "Group of Eight",
        "children": [categories["G7"]["children"][0] + ["RUS"]],
    }
    categories["G20"] = {
        "title": "Group of 20",
        "children": [
            categories["G8"]["children"][0]
            + [
                "ARG",
                "AUS",
                "BRA",
                "CHN",
                "IND",
                "IDN",
                "MEX",
                "SAU",
                "ZAF",
                "KOR",
                "TUR",
            ]
        ],
    }
    return categories


def add_unfccc_names(categories):
    categories["BOL"]["info"]["unfccc_name"] = "Bolivia (Plurinational State of)"
    categories["COD"]["info"] = {"unfccc_name": "Democratic Republic of the Congo"}
    categories["VAT"]["info"] = {"unfccc_name": "Holy See"}
    categories["IRN"]["info"]["unfccc_name"] = "Iran (Islamic Republic of)"
    categories["FSM"]["info"]["unfccc_name"] = "Micronesia (Federated States of)"
    categories["KOR"]["info"]["unfccc_name"] = "Republic of Korea"
    categories["PSE"]["info"]["unfccc_name"] = "State of Palestine"
    categories["VEN"]["info"]["unfccc_name"] = "Venezuela (Bolivarian Republic of)"
    return categories


def add_unfccc_categories(categories):
    categories["Annex-I"] = {
        "title": "Annex-I parties to the UNFCCC",
        "comment": "Parties to the UN Framework Convention on Climate Change "
        "listed in Annex I of the Convention.",
        "alternative_codes": ["ANNEXI"],
        "children": [
            [
                "AUS",
                "AUT",
                "BLR",
                "BEL",
                "BGR",
                "CAN",
                "HRV",
                "CYP",
                "CZE",
                "DNK",
                "EST",
                "FIN",
                "FRA",
                "DEU",
                "GRC",
                "HUN",
                "ISL",
                "IRL",
                "ITA",
                "JPN",
                "LVA",
                "LIE",
                "LTU",
                "LUX",
                "MLT",
                "MCO",
                "NLD",
                "NZL",
                "NOR",
                "POL",
                "PRT",
                "ROU",
                "RUS",
                "SVK",
                "SVN",
                "ESP",
                "SWE",
                "CHE",
                "TUR",
                "UKR",
                "GBR",
                "USA",
                "EU",
            ]
        ],
    }
    categories["Non-Annex-I"] = {
        "title": "Non-Annex-I parties to the UNFCCC",
        "comment": "Parties to the UN Framework Convention on Climate Change "
        "not listed in Annex I of the Convention.",
        "alternative_codes": ["NONANNEXI", "Non Annex-I"],
        "children": [
            [
                "AFG",
                "ALB",
                "DZA",
                "AND",
                "AGO",
                "ATG",
                "ARG",
                "ARM",
                "AZE",
                "BHS",
                "BHR",
                "BGD",
                "BRB",
                "BLZ",
                "BEN",
                "BTN",
                "BIH",
                "BWA",
                "BRA",
                "BRN",
                "BFA",
                "BDI",
                "CPV",
                "KHM",
                "CMR",
                "CAF",
                "TCD",
                "CHL",
                "CHN",
                "COL",
                "COM",
                "COG",
                "COK",
                "CRI",
                "CIV",
                "CUB",
                "DJI",
                "DMA",
                "DOM",
                "ECU",
                "EGY",
                "SLV",
                "GNQ",
                "ERI",
                "SWZ",
                "ETH",
                "FJI",
                "GAB",
                "GMB",
                "GEO",
                "GHA",
                "GRD",
                "GTM",
                "GIN",
                "GNB",
                "GUY",
                "HTI",
                "HND",
                "IND",
                "IDN",
                "IRQ",
                "ISR",
                "JAM",
                "JOR",
                "KAZ",
                "KEN",
                "KIR",
                "KWT",
                "KGZ",
                "LAO",
                "LBN",
                "LSO",
                "LBR",
                "LBY",
                "MDG",
                "MWI",
                "MYS",
                "MDV",
                "MLI",
                "MHL",
                "MRT",
                "MUS",
                "MEX",
                "MNG",
                "MNE",
                "MAR",
                "MOZ",
                "MMR",
                "NAM",
                "NRU",
                "NPL",
                "NIC",
                "NER",
                "NGA",
                "NIU",
                "MKD",
                "OMN",
                "PAK",
                "PLW",
                "PAN",
                "PNG",
                "PRY",
                "PER",
                "PHL",
                "QAT",
                "RWA",
                "KNA",
                "LCA",
                "VCT",
                "WSM",
                "SMR",
                "STP",
                "SAU",
                "SEN",
                "SRB",
                "SYC",
                "SLE",
                "SGP",
                "SLB",
                "SOM",
                "ZAF",
                "SSD",
                "LKA",
                "SDN",
                "SUR",
                "SYR",
                "TJK",
                "THA",
                "TLS",
                "TGO",
                "TON",
                "TTO",
                "TUN",
                "TKM",
                "TUV",
                "UGA",
                "ARE",
                "URY",
                "UZB",
                "VUT",
                "VNM",
                "YEM",
                "ZMB",
                "ZWE",
                "PRK",
                "COD",
                "VAT",
                "IRN",
                "FSM",
                "KOR",
                "MDA",
                "PSE",
                "TZA",
                "VEN",
                "BOL",
            ]
        ],
    }
    categories["UNFCCC"] = {
        "title": "Parties to the UNFCCC",
        "comment": "Parties to the UN Framework Convention on Climate Change.",
        "children": [
            categories["Annex-I"]["children"][0]
            + categories["Non-Annex-I"]["children"][0],
            ["Annex-I", "Non-Annex-I"],
        ],
    }

    return categories


def add_eu_categories(categories):
    categories["EU_1993"] = {
        "title": "European Union from 1993 to 1994",
        "comment": "The European Union from 1993-11-1 to 1994-12-31.",
        "alternative_codes": ["EU12", "EU-12"],
        "children": [
            [
                "BEL",
                "DNK",
                "FRA",
                "DEU",
                "GRC",
                "IRL",
                "ITA",
                "LUX",
                "NLD",
                "PRT",
                "ESP",
                "GBR",
            ]
        ],
    }
    categories["EU_1995"] = {
        "title": "European Union from 1995 to 2004",
        "comment": "The European Union from 1995-01-01 to 2004-04-30.",
        "alternative_codes": ["EU15", "EU-15"],
        "children": [categories["EU_1993"]["children"][0] + ["AUT", "FIN", "SWE"]],
    }
    categories["EU_2004"] = {
        "title": "European Union from 2004 to 2006",
        "comment": "The European Union from 2004-05-01 to 2006-12-31.",
        "alternative_codes": ["EU25", "EU-25"],
        "children": [
            categories["EU_1995"]["children"][0]
            + ["CYP", "CZE", "EST", "HUN", "LVA", "LTU", "MLT", "POL", "SVK", "SVN"]
        ],
    }
    # Do not include EU27 as it is unclear what it means
    categories["EU_2007"] = {
        "title": "European Union from 2007 to 2013",
        "comment": "The European Union from 2007-01-01 to 2013-06-30.",
        "alternative_codes": ["EU27_2007", "EU-27_2007"],
        "children": [categories["EU_2004"]["children"][0] + ["BGR", "ROU"]],
    }
    categories["EU_2013"] = {
        "title": "European Union from 2013 to 2020",
        "comment": "The European Union from 2013-07-01 to 2020-01-31.",
        "alternative_codes": ["EU28", "EU-28"],
        "children": [categories["EU_2007"]["children"][0] + ["HRV"]],
    }
    eu2020_children = categories["EU_2013"]["children"][0].copy()
    eu2020_children.remove("GBR")
    # Do not include EU27 as it is unclear what it means
    categories["EU_2020"] = {
        "title": "European Union",
        "comment": "The European Union since 2020-02-01 to date. Note that the 'EU' "
        "code will always refer to the current EU, use EU_2020 if you need "
        "a stable code.",
        "alternative_codes": ["EU27_2020", "EU-27_2020", "EU", "EU27BX"],
        "children": [eu2020_children],
    }

    return categories


def load_countries() -> (
    dict[str, typing.Union[str, dict[str, str], list[str], list[list[str]]]]
):
    """Load countries from the iso-codes debian package."""
    r = requests.get(
        URL,
        headers={"Accept": "application/json"},
    )

    categories: dict[str, typing.Union[str, dict[str, str]]] = {}
    for country in r.json()["3166-1"]:
        country_spec = {
            "title": country["name"],
            "alternative_codes": [country["alpha_2"], country["numeric"]],
        }

        info = {}
        if "official_name" in country:
            info["official_name"] = country["official_name"]
        if "common_name" in country:
            info["common_name"] = country["common_name"]
        if info:
            country_spec["info"] = info

        categories[country["alpha_3"]] = country_spec

    return categories


if __name__ == "__main__":
    ISO3 = main()

    ISO3.to_yaml(OUTPATH)

    climate_categories.HierarchicalCategorization.from_yaml(OUTPATH)



"""Tests for _conversions"""

import datetime
import importlib
import importlib.resources
from io import StringIO

import pytest
import strictyaml as sy

import climate_categories
import climate_categories._conversions as conversions
import climate_categories.tests
import climate_categories.tests.data


class TestConversionRuleSpec:
    def test_to_csv_row(self):
        crs = conversions.ConversionRuleSpec(
            factors_categories_a={"A": 1, "B": -1},
            factors_categories_b={"C": 1},
            auxiliary_categories={"aux": {"auxB", "auxA"}},
            comment="nice comment",
        )
        assert crs.to_csv_row() == [
            "A - B",
            "auxA auxB",
            "C",
            "nice comment",
        ]

    def test_str(self):
        crs = conversions.ConversionRuleSpec(
            factors_categories_a={"A": 1, "B": -1},
            factors_categories_b={"C": 1},
            auxiliary_categories={"aux": set()},
            comment="nice comment",
        )
        assert str(crs) == "A - B,,C,nice comment"
        crs_csv = conversions.ConversionRuleSpec(
            factors_categories_a={"A": 1, "B": -1},
            factors_categories_b={"C": 1},
            auxiliary_categories={"aux": {"auxB", "auxA"}},
            comment="nice comment",
            csv_original_text='A-B, auxB auxA, "C",nice comment',
            csv_line_number=4,
        )
        assert str(crs_csv) == 'A-B, auxB auxA, "C",nice comment'


class TestConversionRule:
    def test_str(self):
        C96 = climate_categories.IPCC1996
        C06 = climate_categories.IPCC2006
        gas = climate_categories.gas
        cr = conversions.ConversionRule(
            factors_categories_a={C96["1"]: 1, C96["2"]: -1},
            factors_categories_b={C06["3"]: 1},
            auxiliary_categories={gas: set()},
            comment="nice comment",
        )
        assert str(cr) == "1 - 2,,3,nice comment"

        reparsed = conversions.ConversionRuleSpec.from_csv_row(
            str(cr).split(","), ["gas"]
        ).hydrate(C96, C06, {"gas": gas})
        assert reparsed == cr

        cr_csv = conversions.ConversionRule(
            factors_categories_a={C96["1"]: 1, C96["2"]: -1},
            factors_categories_b={C06["3"]: 1},
            auxiliary_categories={gas: {gas["CO2"]}},
            comment="nice comment",
            csv_original_text='"1"-"2", CO2 , "3",nice comment',
            csv_line_number=4,
        )
        assert str(cr_csv) == '"1"-"2", CO2 , "3",nice comment'

    def test_format_with_lineno(self):
        C96 = climate_categories.IPCC1996
        C06 = climate_categories.IPCC2006
        gas = climate_categories.gas
        cr = conversions.ConversionRule(
            factors_categories_a={C96["1"]: 1, C96["2"]: -1},
            factors_categories_b={C06["3"]: 1},
            auxiliary_categories={gas: set()},
            comment="nice comment",
            csv_line_number=4,
        )
        assert cr.format_with_lineno() == "<Rule '1 - 2,,3,nice comment' from line 4>"

    def test_cardinality(self):
        C96 = climate_categories.IPCC1996
        C06 = climate_categories.IPCC2006
        cr = conversions.ConversionRule(
            factors_categories_a={C96["1"]: 1, C96["2"]: -1},
            factors_categories_b={C06["3"]: 1},
            auxiliary_categories={},
            comment="",
            csv_line_number=4,
        )
        assert cr.cardinality_a == "many"
        assert cr.cardinality_b == "one"

    def test_restricted(self):
        C96 = climate_categories.IPCC1996
        C06 = climate_categories.IPCC2006
        gas = climate_categories.gas
        unrestricted = conversions.ConversionRule(
            factors_categories_a={C96["1"]: 1, C96["2"]: -1},
            factors_categories_b={C06["3"]: 1},
            auxiliary_categories={},
            comment="",
            csv_line_number=4,
        )
        restricted = conversions.ConversionRule(
            factors_categories_a={C96["1"]: 1, C96["2"]: -1},
            factors_categories_b={C06["3"]: 1},
            auxiliary_categories={gas: set("NO2")},
            comment="",
            csv_line_number=4,
        )
        assert not unrestricted.is_restricted
        assert restricted.is_restricted


class TestConversionSpec:
    def test_good_csv(self, tmp_path):
        # write CSV resource to file to test also file opening path
        good_csv = (
            importlib.resources.files("climate_categories.tests.data")
            .joinpath("good_conversion.csv")
            .open()
        )
        temp_csv = tmp_path / "gc.csv"
        with temp_csv.open("w") as fd2:
            fd2.write(good_csv.read())

        # Now actually read from the file
        conv = conversions.ConversionSpec.from_csv(temp_csv)
        assert conv.categorization_a_name == "A"
        assert conv.categorization_b_name == "B"
        assert conv.comment == "A correct conversion specification file"
        assert conv.version == "1.2.3.4"
        assert conv.references == "expert judgement"
        assert conv.last_update == datetime.datetime(2099, 12, 31, 0, 0)
        assert conv.auxiliary_categorizations_names == ["aux1", "aux2"]
        assert len(conv.rule_specs) == 9
        assert conv.rule_specs[0] == conversions.ConversionRuleSpec(
            factors_categories_a={"asdf": 1, "fdsa": 1},
            factors_categories_b={"asdf": 1},
            auxiliary_categories={"aux1": set(), "aux2": set()},
            csv_line_number=6,
            csv_original_text="asdf + fdsa,,,asdf",
        )
        assert conv.rule_specs[1] == conversions.ConversionRuleSpec(
            factors_categories_a={"A.5": 1},
            factors_categories_b={"4": 1},
            auxiliary_categories={
                "aux1": {"3", "4", "A", "argl-5"},
                "aux2": {"B A", "A", "B"},
            },
            csv_line_number=7,
            csv_original_text='A.5,3 4 A "argl-5", "B A" B A,4',
        )
        assert conv.rule_specs[2] == conversions.ConversionRuleSpec(
            factors_categories_a={"b": 1, "argl.5": 1, "c": 1},
            factors_categories_b={"D": 1},
            auxiliary_categories={"aux1": set(), "aux2": set()},
            comment="nobody needs argl",
            csv_line_number=8,
            csv_original_text='b + "argl.5" + c,,,D,nobody needs argl',
        )
        assert conv.rule_specs[3] == conversions.ConversionRuleSpec(
            factors_categories_a={"b": 1, "argl,5": 1, "c": 1},
            factors_categories_b={"D": 1},
            auxiliary_categories={"aux1": set(), "aux2": set()},
            csv_line_number=9,
            csv_original_text='b + "argl,5" + c,,,D',
        )
        assert conv.rule_specs[4] == conversions.ConversionRuleSpec(
            factors_categories_a={"b": 1, 'argl"5': 1, "c": 1},
            factors_categories_b={"D": 1},
            auxiliary_categories={"aux1": set(), "aux2": set()},
            csv_line_number=10,
            csv_original_text='b + "argl\\"5" + c,,,D',
        )
        assert conv.rule_specs[5] == conversions.ConversionRuleSpec(
            factors_categories_a={"b": 1, 'argl"5': 1, "c": 1},
            factors_categories_b={"D": 1, "E": 1},
            auxiliary_categories={"aux1": set(), "aux2": set()},
            csv_line_number=11,
            csv_original_text='b + "argl\\"5" + c,,,D+E',
        )
        assert conv.rule_specs[6] == conversions.ConversionRuleSpec(
            factors_categories_a={"argl,5": 1},
            factors_categories_b={"D": 1},
            auxiliary_categories={"aux1": set(), "aux2": set()},
            csv_line_number=12,
            csv_original_text='"argl,5",,,D',
        )
        assert conv.rule_specs[7] == conversions.ConversionRuleSpec(
            factors_categories_a={"+": 1, "-": -1},
            factors_categories_b={"-": 1},
            auxiliary_categories={"aux1": set(), "aux2": set()},
            csv_line_number=13,
            csv_original_text='"+" - "-",,,"-"',
        )
        assert conv.rule_specs[8] == conversions.ConversionRuleSpec(
            factors_categories_a={"b": 1},
            factors_categories_b={"asdf": 1, "4": 1},
            auxiliary_categories={"aux1": set(), "aux2": set()},
            csv_line_number=14,
            csv_original_text="b,,,asdf+4",
        )

        assert repr(conv) == "<ConversionSpec 'A' <-> 'B' with 9 rules>"

    def test_formula_broken(self):
        csv = StringIO(
            "# comment: no comment\n"
            "A,species,B,comment\n"
            "A.1+,something,C,broken formula\n"
        )
        with pytest.raises(ValueError, match="line 3.*Could not parse"):
            climate_categories._conversions.ConversionSpec.from_csv(csv)

    def test_unknown_meta_data(self):
        csv = StringIO(
            "# comment: no comment\n"
            "# interjection: What you guys are referring to as Linux\n"
        )
        with pytest.raises(sy.exceptions.YAMLValidationError):
            climate_categories._conversions.ConversionSpec.from_csv(csv)

    def test_comment_missing(self):
        csv = StringIO("# comment: no comment\n" "A,aux,B\n" "A.1,D,A.2\n")
        with pytest.raises(
            ValueError, match="Last column must be 'comment', but isn't."
        ):
            climate_categories._conversions.ConversionSpec.from_csv(csv)


def load_conversion_from_csv(fname: str):
    fd = (
        importlib.resources.files("climate_categories.tests.data")
        .joinpath(fname)
        .open()
    )
    gc = conversions.ConversionSpec.from_csv(fd)

    cats = {
        cat_name: climate_categories.Categorization.from_yaml(
            importlib.resources.files("climate_categories.tests.data")
            .joinpath(f"good_conversion_{cat_name}.yaml")
            .open()
        )
        for cat_name in ("A", "B", "aux1", "aux2")
    }

    return gc.hydrate(cats)


@pytest.fixture
def good_conversion():
    return load_conversion_from_csv("good_conversion.csv")


@pytest.fixture
def good_conversion_reversed():
    return load_conversion_from_csv("good_conversion_reversed.csv")


class TestConversion:
    def test_not_implemented(self):
        with pytest.raises(NotImplementedError):
            climate_categories.IPCC1996.conversion_to(climate_categories.gas)

    def test_symmetry(self):
        a = climate_categories.IPCC1996.conversion_to(climate_categories.IPCC2006)
        b = climate_categories.IPCC2006.conversion_to(climate_categories.IPCC1996)
        assert a == b.reversed()
        assert a == a.reversed().reversed()

    def test_str_arg(self):
        a = climate_categories.IPCC1996.conversion_to(climate_categories.IPCC2006)
        b = climate_categories.IPCC1996.conversion_to("IPCC2006")
        assert a == b

    def test_find_unmapped_categories(self, good_conversion: conversions.Conversion):
        missing_a, missing_b = good_conversion.find_unmapped_categories()
        assert missing_a == {good_conversion.categorization_a["unmapped"]}
        assert missing_b == set()

    def test_reverse(
        self,
        good_conversion: conversions.Conversion,
        good_conversion_reversed: conversions.Conversion,
    ):
        assert good_conversion.reversed() == good_conversion_reversed
        assert good_conversion_reversed.reversed() == good_conversion

    def test_repr(self, good_conversion: conversions.Conversion):
        assert repr(good_conversion) == "<Conversion 'A' <-> 'B' with 9 rules>"

    def test_not_allowed(self):
        with pytest.raises(ValueError, match="Error in line 10: Could not parse:"):
            load_conversion_from_csv("broken_conversion_not_allowed.csv")

    def test_not_existing(self):
        with pytest.raises(
            ValueError, match="Error in line 10: 'notexisting' not in A"
        ):
            load_conversion_from_csv("broken_conversion_not_existing.csv")

    def test_describe(self, good_conversion: conversions.Conversion):
        assert (
            good_conversion.describe_detailed()
            == """# Mapping between A and B

## Simple direct mappings

Only for aux1 in ['3', '4', 'A', 'argl-5'] and aux2 in ['A', 'B', 'B A']
A A.5 The category A.5, aka b
B 4 the number four

A argl,5 weirder characters
B D d


## One-to-many mappings - one A to many B

A A.5 The category A.5, aka b
⮁
B asdf ASDF
B 4 the number four


## Many-to-one mappings - many A to one B

A asdf ASDF
A fdsa FDSA
⮁
B asdf ASDF

A A.5 The category A.5, aka b
A argl.5 We are testing weird characters
A c The sea
⮁
B D d
# Comment: 'nobody needs argl'

A A.5 The category A.5, aka b
A argl,5 weirder characters
A c The sea
⮁
B D d

A A.5 The category A.5, aka b
A argl"5 stretching the dredibility
A c The sea
⮁
B D d

A + sure, make the code for your category a plus sign
-1 * A - even better
⮁
B - a minus sign


## Many-to-many mappings - many A to many B

A A.5 The category A.5, aka b
A argl"5 stretching the dredibility
A c The sea
⮁
B D d
B E e


## Unmapped categories

### A
unmapped it hurts to be forgotten

### B


"""
        )


def test_over_counting_problem():
    C96 = climate_categories.IPCC1996
    C06 = climate_categories.IPCC2006
    OP = conversions.OverCountingProblem(
        category=C96["1"],
        leave_node_groups=[{C06["1"], C06["2"]}],
        rules=[
            conversions.ConversionRule(
                factors_categories_a={C96["1"]: 1},
                factors_categories_b={C06["1"]: 1, C06["2"]: 1},
                auxiliary_categories={},
            )
        ],
    )

    assert (
        str(OP)
        == """<IPCC1996: '1'> is possibly counted multiple times
involved leave groups categories: [[<IPCC2006: '1'>, <IPCC2006: '2'>]]
involved rules: <Rule '1,1 + 2,'>."""
    )


def test_relevant_rules():
    C96 = climate_categories.IPCC1996
    C06 = climate_categories.IPCC2006

    conv = C96.conversion_to(C06)

    assert conv.relevant_rules({C96["1"], C96["2"]}) == conv.relevant_rules(
        {C96["1"], C96["2"]}, source_categorization=C96
    )

    assert len(conv.relevant_rules({C96["4.D"]})) == 1
    assert len(conv.relevant_rules({C96["4.B.10"]}, simple_sums_only=True)) == 1
    assert conv.relevant_rules(set()) == []


def get_test_data_filepath(fname: str):
    return importlib.resources.files("climate_categories.tests.data").joinpath(fname)


def test_read_csv_in_conversion_class():
    fd = get_test_data_filepath("example_conversion.IPCC1996.IPCC2006.csv")

    conv_from_conversion_spec = conversions.ConversionSpec.from_csv(fd)
    conv_from_conversion_spec = conv_from_conversion_spec.hydrate(
        cats=climate_categories.cats
    )

    conv_from_conversion = conversions.Conversion.from_csv(fd)

    assert vars(conv_from_conversion_spec) == vars(conv_from_conversion)


def test_read_conversion_from_csv_with_custom_categorizations():
    categorisation_a = climate_categories.from_yaml(
        get_test_data_filepath("simple_categorisation_a.yaml")
    )

    categorisation_b = climate_categories.from_yaml(
        get_test_data_filepath("simple_categorisation_b.yaml")
    )

    cats = {"A": categorisation_a, "B": categorisation_b}

    conv = climate_categories.Conversion.from_csv(
        get_test_data_filepath("simple_conversion.csv"), cats=cats
    )

    assert conv.categorization_a_name == "A"
    assert conv.categorization_b_name == "B"


def test_read_conversion_from_csv_with_existing_categorizations():
    conv = climate_categories.Conversion.from_csv(
        get_test_data_filepath("test_conversion_with_existing_categorizations.csv")
    )

    assert conv.categorization_a_name == "BURDI"
    assert conv.categorization_b_name == "IPCC2006_PRIMAP"


def test_read_conversion_from_csv_with_existing_categorizations_aux_dims():
    conv = climate_categories.Conversion.from_csv(
        get_test_data_filepath(
            "test_conversion_with_existing_categorizations_aux_dims.csv"
        )
    )

    assert conv.categorization_a_name == "BURDI"
    assert conv.categorization_b_name == "IPCC2006_PRIMAP"
    assert conv.auxiliary_categorizations_names == ["gas"]



"""Run this via `make climate_categories/data/CRFDI_class.yaml` in the main directory."""

import datetime
import pathlib

import numpy as np
import treelib
import unfccc_di_api

import climate_categories

OUTPATH = pathlib.Path("./climate_categories/data/CRFDI_class.yaml")


def parse_classifications(rao):
    new_categories = {}
    new_children = []
    new_alternative_codes = {}
    for parent_category in climate_categories.CRFDI.values():
        classification_ids = np.unique(
            rao.variables.loc[
                rao.variables["categoryId"].isin(
                    [int(x) for x in parent_category.info["numerical_ids"]]
                ),
                "classificationId",
            ]
        )

        new_children_for_category = []
        i = 0
        for cid in sorted(classification_ids):
            if cid == 10510:  # Total for category, i.e. not a sub-category
                # Just add additional altcodes
                primary_altcode = f"{parent_category.codes[0]}-{cid}"
                new_alternative_codes[primary_altcode] = parent_category.codes[0]
                for nid in parent_category.info["numerical_ids"]:
                    altcode = f"{nid}-{cid}"
                    if altcode != primary_altcode:
                        new_alternative_codes[altcode] = parent_category.codes[0]
                continue
            i += 1
            code = f"{parent_category.codes[0]}-{i}"
            altcodes = [f"{parent_category.codes[0]}-{cid}"]
            numerical_altcodes = [
                f"{nid}-{cid}" for nid in parent_category.info["numerical_ids"]
            ]
            for numerical_altcode in numerical_altcodes:
                if numerical_altcode not in altcodes:
                    altcodes.append(numerical_altcode)

            new_categories[code] = {
                "title": rao.classifications.loc[cid, "name"],
                "alternative_codes": altcodes,
            }
            new_children_for_category.append(code)

        if new_children_for_category:
            new_children.append((parent_category.codes[0], new_children_for_category))

    return new_categories, new_children, new_alternative_codes


def parse_refrigerant_measures(rao: unfccc_di_api.UNFCCCSingleCategoryApiReader):
    refrigerant_emission_measures = {
        "Emissions from manufacturing": 1,
        "Emissions from stocks": 2,
        "Emissions from disposal": 3,
    }
    new_categories = {}
    new_children = []
    new_alternative_codes = {}
    for parent_category in climate_categories.CRFDI.values():
        measure_ids = np.unique(
            rao.variables.loc[
                rao.variables["categoryId"].isin(
                    [int(x) for x in parent_category.info["numerical_ids"]]
                ),
                "measureId",
            ]
        )

        new_children_for_category = []
        for mid in sorted(measure_ids):
            try:
                measure = rao.measure_tree[mid].tag
            except treelib.tree.NodeIDAbsentError:
                continue
            if measure not in refrigerant_emission_measures:
                continue
            short_mid = refrigerant_emission_measures[measure]
            code = f"{parent_category.codes[0]}-m{short_mid}"
            altcodes = [f"{parent_category.codes[0]}-m{mid}"]
            numerical_altcodes = [
                f"{nid}-m{mid}" for nid in parent_category.info["numerical_ids"]
            ]
            for numerical_altcode in numerical_altcodes:
                if numerical_altcode not in altcodes:
                    altcodes.append(numerical_altcode)

            new_categories[code] = {
                "title": measure,
                "alternative_codes": altcodes,
            }
            new_children_for_category.append(code)

        if new_children_for_category:
            new_children.append((parent_category.codes[0], new_children_for_category))

    return new_categories, new_children, new_alternative_codes


def main():
    rao = unfccc_di_api.UNFCCCSingleCategoryApiReader(party_category="annexOne")
    categories, children, alternative_codes = parse_classifications(rao)

    categories_add, children_add, alternative_codes_add = parse_refrigerant_measures(
        rao
    )
    categories.update(categories_add)
    children += children_add
    alternative_codes.update(alternative_codes_add)

    CRFDI_class = climate_categories.CRFDI.extend(
        categories=categories,
        children=children,
        alternative_codes=alternative_codes,
        name="class",
        title=" + classifications",
        comment=" extended by sub-classifications also provided by the DI flexible "
        "query interface",
        last_update=datetime.date.fromisoformat("2021-11-05"),
    )
    CRFDI_class.institution = "UNFCCC"
    CRFDI_class.references = (
        "https://unfccc.int/process-and-meetings/"
        "transparency-and-reporting/greenhouse-gas-data/data-interface-help#eq-7"
    )
    CRFDI_class.total_sum = False  # unfortunately, not generally true anymore

    CRFDI_class.to_yaml(OUTPATH)

    climate_categories.HierarchicalCategorization.from_yaml(OUTPATH)


if __name__ == "__main__":
    main()
